import base64
import json
import logging
from typing import Annotated, Literal, Union
from urllib import parse

from pydantic import BaseModel, Field, SecretStr
from pydantic.functional_validators import model_validator
from requests import Request, Session
from requests.adapters import HTTPAdapter, Retry

from ..__types import DictData
from ..models import MetricData
from .__abc import BaseMetric

logger = logging.getLogger("jett")


class BasicAuth(BaseModel):
    """Basic Authenticate."""

    type: Literal["basic"]
    username: str
    password: SecretStr

    def apply(self, request: Request) -> Request:
        """Apply Basic Authentication."""
        # ARCHIVE:
        # from requests.auth import HTTPBasicAuth
        #
        # request.auth = HTTPBasicAuth(self.username, self.password)
        encoded = base64.b64encode(
            bytes(
                f"{self.username}:{self.password.get_secret_value()}", "utf-8"
            )
        )
        request.headers.update({"Authorization": f"Basic {encoded.decode()}"})
        return request


RestAPIAuth = Annotated[
    Union[BasicAuth,],
    Field(discriminator="type"),
]


class RestAPI(BaseMetric):
    """Metric RestAPI model."""

    type: Literal["restapi"]
    base_url: str
    path: str
    method: Literal["POST", "GET", "PUT", "PATCH"]
    header: dict[str, str | int] = Field(default_factory=dict)
    authentication: RestAPIAuth | None = None
    retry: int = 0
    retry_on_status: list[int] = Field(default_factory=list)
    timeout: int = 30

    @model_validator(mode="after")
    def __post_validate(self):
        """Validate URL."""
        parse_result = parse.urlparse(f"{self.base_url}{self.path}")
        if (
            parse_result.scheme not in ["http", "https"]
            or parse_result.netloc == ""
        ):
            raise ValueError("invalid base_url or path")

        if self.retry > 0 and len(self.retry_on_status) == 0:
            raise ValueError(
                "Please specify response status code for retrying mechanism"
            )
        return self

    def attach_retry(self, session: Session) -> None:
        """Attach Retry mechanism to the Session object."""
        retry_mec = Retry(
            total=self.retry,
            backoff_factor=2,
            status_forcelist=self.retry_on_status,
            allowed_methods=[self.method],
        )
        session.mount("http://", HTTPAdapter(max_retries=retry_mec))
        session.mount("https://", HTTPAdapter(max_retries=retry_mec))

    def emit(self, data: MetricData, convert: DictData, **kwargs) -> None:
        """Emit"""
        with Session() as session:
            url: str = f"{self.base_url}{self.path}"
            request = Request(
                method=self.method,
                url=url,
                data=json.dumps(convert, default=str),
                headers=self.header,
            )

            if self.retry > 0 and len(self.retry_on_status) > 0:
                self.attach_retry(session)

            if self.authentication:
                self.authentication.apply(request)

            logger.info("📩 Metric - RestAPI")
            logger.info(
                f"Pushing metric to: {self.method} {url}, with payload:\n"
                f"{json.dumps(convert, default=str, indent=1)}"
            )
            response = session.send(request.prepare(), timeout=self.timeout)
            response.raise_for_status()
            logger.info(
                f"Response status code of pushing metric to: "
                f"{self.method} {url}, is {response.status_code}",
            )



from __future__ import annotations

from typing import Literal

try:
    import daft
    from daft import DataFrame

    DAFT_EXISTS: bool = True
except ImportError:
    DAFT_EXISTS: bool = False

from jett.__types import DictData
from jett.engine.__abc import BaseSource
from jett.models import MetricSource, Shape


class LocalJsonFile(BaseSource):
    type: Literal["local"]
    file_format: Literal["json"]
    path: str

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[DataFrame, Shape]:
        import daft

        df: DataFrame = daft.read_json(
            path=self.path,
            file_path_column=None,
        )
        return df, Shape()

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalCsvFile(BaseSource):
    type: Literal["local"]
    file_format: Literal["csv"]
    path: str
    delimiter: str | None = None
    header: bool = True

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[DataFrame, Shape]:
        if not DAFT_EXISTS:
            raise ModuleNotFoundError("Does not install daft yet.")

        df: DataFrame = daft.read_csv(
            path=self.path,
            file_path_column=None,
            delimiter=self.delimiter,
            has_headers=self.header,
        )
        return df, Shape()

    def inlet(self) -> tuple[str, str]:
        return "local", self.path



import logging
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from polars import DataFrame

from pydantic import Field
from pydantic.functional_validators import field_validator

from ...__types import DictData
from ...models import ColDetail, Context, MetricEngine, MetricTransform, Result
from ..__abc import BaseEngine
from .sink import Sink
from .source import Source
from .transform import Transform

logger = logging.getLogger("jett")


class Polars(BaseEngine):
    """Polars Engine model."""

    type: Literal["polars"] = Field(description="A type of engine.")
    source: Source
    transforms: list[Transform] = Field(default_factory=list)
    sink: list[Sink] = Field(description="A list of Sink models.")

    @field_validator(
        "sink",
        mode="before",
        json_schema_input_type=Sink | list[Sink],
    )
    def __prepare_sink(cls, data: Any) -> Any:
        """Prepare the sink field value that should be list of Sink model."""
        return [data] if not isinstance(data, list) else data

    def execute(
        self,
        context: Context,
        engine: DictData,
        metric: MetricEngine,
    ) -> "DataFrame":
        """Execute Polars engine method.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricEngine): A metric engine that was set from handler
                step for passing custom metric data.

        Returns:
            DataFrame:
        """
        logger.info("🏗️ Start execute with Polars engine.")
        # NOTE: Start run source handler.
        df: DataFrame = self.source.handle_load(context, engine=engine)

        # NOTE: Start run transform handler.
        df: DataFrame = self.handle_apply(df, context, engine=engine)

        # NOTE: Start multi-sink by sequential strategy.
        for _sink in self.sink:
            _sink.handle_save(df, context, engine=engine)

        return df

    def apply(
        self,
        df: "DataFrame",
        context: Context,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> "DataFrame":
        """Apply Polars engine transformation to the source. This method
        will apply all operators.

        Args:
            df (DataFrame): A Polars DataFrame.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricTransform): A metric transform that was set from
                handler step for passing custom metric data.
        """
        for t in self.transforms:
            df: DataFrame = t.handle_apply(df, context, engine=engine)
        return df

    def set_result(self, df: "DataFrame", context: Context) -> Result:
        """Set the Result object from executed Polars DataFrame.

        Args:
            df (DataFrame): A Polars DataFrame.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
        """
        return Result(
            data=[],
            columns=[
                ColDetail(name=name, dtype=str(dtype))
                for name, dtype in df.schema.items()
            ],
            schema_dict=df.schema,
        )

    def set_engine_context(self, context: Context, **kwargs) -> DictData:
        """Set Polars engine context. It will set itself to engine context data.

        Returns:
            A mapping of the engine context data:
                - engine: Self
        """
        return {"engine": self}



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from ..__types import DictData
from ..models import ALWAYS, EmitCond, MetricData
from .convertor import Convertor


class BaseMetric(BaseModel, ABC):
    """Base Metric abstract model."""

    model_config = ConfigDict(use_enum_values=True)

    type: str = Field(description="A metric type.")
    custom_metric: dict[str, Any] = Field(
        default_factory=dict, description="A custom metric."
    )
    convertor: Convertor = Field(
        default="basic",
        description=(
            "A convertor type that want to prepare metric data before using "
            "on the implemented emit method."
        ),
    )
    condition: EmitCond = Field(
        default=ALWAYS,
        description=(
            "An emitter condition to allow call the emit method after "
            "execution finish."
        ),
    )

    @abstractmethod
    def emit(self, data: MetricData, convert: DictData, **kwargs) -> None:
        """Emit abstraction method for implement egress logic that send the
        metric data after engine execution end.

        Args:
            data (MetricData): A MetricData model that passing after end the
                engine execution.
            convert (DictData): A converted data from the metric data.
        """

    def handle_emit(self, data: MetricData, **kwargs) -> None:
        """Handle Emit method by passing converted data by the convert method.

        Args:
            data (MetricData): A MetricData model instance to pass to emit
                and convert methods.
        """
        covert_data: DictData = self.convert(data)
        self.emit(data=data, convert=covert_data, **kwargs)

    def convert(self, data: MetricData) -> DictData:
        """Convert MetricData to the dict object with specific convertor
        function.

        Args:
            data (MetricData): A MetricData model instance to pass to convert
                methods.
        """
        from .convertor import CONVERTOR_REGISTRY

        if self.convertor not in CONVERTOR_REGISTRY:  # pragma: no cov
            raise NotImplementedError(
                f"Convertor type: {self.convertor!r} does not support."
            )
        return CONVERTOR_REGISTRY[self.convertor](
            data, self.custom_metric
        ).convert()



from typing import Annotated, Union

from pydantic import Field

from .files.local import LocalCsvFile, LocalJsonFile
from .files.s3 import S3CSVFile

LocalFile = Annotated[
    Union[
        LocalCsvFile,
        LocalJsonFile,
    ],
    Field(discriminator="file_format"),
]

S3File = Annotated[
    Union[S3CSVFile,],
    Field(discriminator="file_format"),
]

Source = Annotated[
    Union[
        LocalFile,
        S3File,
    ],
    Field(discriminator="type"),
]



from __future__ import annotations

from pathlib import Path
from typing import Any, Union

DictData = dict[
    Union[str, int],  # NOTE: Key should be str or int
    Union[
        str, int, bool, float, list[Any], dict[Any, Any], Any
    ],  # NOTE: Value that catch from YAML parser.
]
StrOrPath = str | Path
StrOrNone = str | None
TupleStr = tuple[str, ...]
PairStr = tuple[str, str]
PrimitiveType = bool | float | int | str



from typing import Annotated, Union

from pydantic import Field

from .functions import (
    DropColumns,
    ExcludeColumns,
    RenameColumns,
    RenameSnakeCase,
    Sql,
)

Transform = Annotated[
    Union[
        Sql,
        DropColumns,
        ExcludeColumns,
        RenameColumns,
        RenameSnakeCase,
    ],
    Field(discriminator="op"),
]
ListTransform = list[Transform]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from duckdb import DuckDBPyRelation as Relation

import pyarrow as pa
from pydantic import Field

from ...__types import DictData
from ...models import ColDetail, Context, MetricEngine, MetricTransform, Result
from ..__abc import BaseEngine
from .sink import Sink
from .source import Source
from .transform import ListTransform, Transform

logger = logging.getLogger("jett")


class DuckDB(BaseEngine):
    """DuckDB Engine model."""

    type: Literal["duckdb"]
    source: Source = Field(description="A source model.")
    sink: Sink = Field(description="A sink model.")
    transforms: ListTransform = Field(
        default_factory=list,
        description="A list of transform model.",
    )

    def set_engine_context(self, context: Context, **kwargs) -> Any:
        from duckdb.experimental.spark.sql import SparkSession

        return {
            "engine": self,
            "spark": SparkSession.builder.getOrCreate(),
        }

    def execute(
        self, context: Context, engine: DictData, metric: MetricEngine
    ) -> Any:
        """Execute the DuckDB engine.

        Returns:
            Result: A result.
        """
        logger.info("Start execute with DuckDB engine.")
        df: Relation = self.source.handle_load(context, engine=engine)
        df: Relation = self.handle_apply(df, context, engine=engine)
        self.sink.handle_save(df, context, engine=engine)
        return df

    def set_result(self, df: Relation, context: Context) -> Result:
        """Set the Result object for this DuckDB engine.

        Returns:
            Result: A result object.
        """
        arrow_table: pa.Table = df.to_arrow_table()
        schema: pa.Schema = arrow_table.schema
        return Result(
            data=arrow_table.to_pylist(),
            columns=[ColDetail(name=f.name, dtype=str(f.type)) for f in schema],
            schema_dict={f.name: f.type for f in schema},
        )

    def apply(
        self,
        df: Relation,
        context: DictData,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> Relation:
        """Apply transform to the source.

        Args:
            df: Relation
            context: DictData
            engine:
            metric:
        """
        import duckdb

        priority: ListTransform = []
        groups: ListTransform = []
        fallback: ListTransform = []

        logger.debug("Prepare transform analyzer ...")
        for t in self.transforms:
            if t.priority == "pre":
                priority.append(t)
            elif t.priority == "group":
                groups.append(t)
            else:
                fallback.append(t)

        logger.info(f"Priority transform count: {len(priority)}")
        for t in priority:
            logger.info(f"Start priority operator: {t.op!r}")
            df: Relation = t.handle_apply(df, context, engine=engine)

        logger.info(f"Group transform count: {len(groups)}")
        if groups:
            statement_group: list[str] = []
            for g in groups:
                statement_group.append(g.handle_apply_group(df))
            stm: str = f"SELECT {', '.join(statement_group)} FROM df"
            logger.info(f"Statement:\n{stm}")
            df: Relation = duckdb.sql(stm)

        logger.info(f"Fallback transform count: {len(fallback)}")
        for t in fallback:
            logger.info(f"Start fallback operator: {t.op!r}")
            df: Relation = t.handle_apply(df, context, engine=engine)

        return df



from typing import Annotated, Union

from pydantic import Field

from .files.local import (
    LocalCsvFileDataset,
    LocalCsvFileTable,
    LocalJsonFileTable,
    LocalParquetFileDataset,
)

LocalCsvFile = Annotated[
    Union[
        LocalCsvFileDataset,
        LocalCsvFileTable,
    ],
    Field(discriminator="arrow_type"),
]


Source = Annotated[
    Union[
        LocalCsvFile,
        LocalJsonFileTable,
        LocalParquetFileDataset,
    ],
    Field(discriminator="file_format"),
]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pydantic import Field

from jett.engine.__abc import BaseSink

if TYPE_CHECKING:
    from polars import DataFrame


class LocalCSVFile(BaseSink):
    """Local file data source."""

    type: Literal["local"]
    file_format: Literal["csv"]
    path: str
    delimiter: str = "|"
    header: bool = Field(default=True)
    sample_records: int | None = 200

    def save(self, df: DataFrame, *, engine, **kwargs) -> None:
        file_format: str = Path(self.path).suffix

        if file_format in (".csv",):
            return

        # df.write_csv(self.path, separator=self.delimiter)

        raise NotImplementedError(
            f"Local file format: {file_format!r} does not support yet."
        )

    def outlet(self) -> tuple[str, str]:
        return "csv", self.dest()

    def dest(self) -> str:
        return self.path



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame

    PairCol = tuple[Column, str]

from ..utils import (
    extract_cols_selectable,
    has_fix_array_index,
    replace_fix_array_index_with_x_index,
)

logger = logging.getLogger("jett")


class ColMap(BaseModel):
    """Column Map model."""

    name: str = Field(description="A new column name.")
    source: str = Field(
        description="A source column statement before alias with alias.",
    )
    dtype: str | None = Field(default=None, description="A data type")
    allow_quote: bool = True

    def get_null_pair(self) -> PairCol:
        """Create a new null column with specific data type.

        Returns:
            PairCol: A pair of Column instance and its alias name.
        """
        from pyspark.sql.functions import lit

        column: Column = lit(None)
        if self.dtype:
            column = column.cast(self.dtype)
        return column, self.name


class RenameColMap(ColMap):
    """Rename Column Map model. This model add useful methods that need to use
    for rename.
    """

    def get_rename_pair(self) -> PairCol:
        from pyspark.sql.functions import col, expr

        column: Column = (
            expr(self.source)
            if has_fix_array_index(self.source)
            else col(self.source)
        )
        if self.dtype:
            column = column.cast(self.dtype)

        return column, self.name

    def get_rename_pair_fix_non_existed_by_null(self, df: DataFrame) -> PairCol:
        _rename_cache_select_cols: list[str] | None = None
        if not _rename_cache_select_cols:
            _rename_cache_select_cols = extract_cols_selectable(df.schema)

        rep_from_col: str = self.source
        if has_fix_array_index(rep_from_col):
            rep_from_col = replace_fix_array_index_with_x_index(rep_from_col)

        if rep_from_col not in _rename_cache_select_cols:
            logger.info(
                f"Fill null on column: {self.name} (dtype {self.dtype}) due to "
                f"column {self.source} not found",
            )
            return self.get_null_pair()
        return self.get_rename_pair()



"""Abstraction module for keeping necessary abstract model class that making
from ABC and BaseModel.

AbstractClasses:
    BaseEngine: A Base abstraction model and class for any implemented engine.
        The engine is the main model for all abstraction model.
    BaseSource: A Base abstraction model and class for any implemented source
        base on its specific engine.
    BaseTransform: A Base abstraction model and class for any implemented
        transform operator base on its specific engine.
    BaseSink: A Base abstraction model and class for any implemented sink
        base on its specific engine.
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Literal, get_args, get_origin

from pydantic import BaseModel, Field
from pydantic.functional_validators import field_validator

from ..__types import DictData
from ..metric import Metric
from ..models import (
    ALWAYS,
    ONLY_FAILED,
    ONLY_SUCCESS,
    Context,
    MetricData,
    MetricEngine,
    MetricOperatorGroup,
    MetricOperatorOrder,
    MetricSink,
    MetricSource,
    MetricTransform,
    Result,
    Shape,
)

logger = logging.getLogger("jett")


class BaseEngine(BaseModel, ABC):
    """Base Engine abstract model for any implemented engine of Tool opterator.
    This model use handler concept.

    AbstractMethods:
        set_engine_context:
        apply:
        execute:
        set_result: Set the Result object from the output of execution.

    Methods:
        handle_execute:
        handle_apply:
        emit:
        post_execute:
    """

    type: str = Field(
        description=(
            "A engine type that will force to set some specific literal string "
            "values."
        )
    )
    name: str = Field(
        description="A name of this engine template session.",
        json_schema_extra={"title": "Name"},
    )
    app_name: str | None = Field(
        default=None,
        description=(
            "An application name that will use on the target engine session if "
            "it need to define."
        ),
    )
    author: list[str] = Field(
        default_factory=list,
        json_schema_extra={
            "title": "Author",
            "description": "An author name of this Tool configuration template",
        },
    )
    owner: list[str] = Field(
        default_factory=list,
        json_schema_extra={
            "title": "Owner",
            "description": "Owner name of this requirement.",
        },
    )
    tags: list[str] = Field(
        default_factory=list,
        description="A list of tag that will help to find a group of template.",
    )
    parent_dir: Path | None = Field(
        default=None,
        description="A parent directory of that configuration location.",
    )
    metrics: list[Metric] = Field(
        default_factory=list,
        description="A list of metric model that want to emit result.",
    )

    @field_validator(
        "author",
        "owner",
        mode="before",
        json_schema_input_type=str | list[str] | None,
    )
    def __convert_str_to_list_of_str(cls, data: Any) -> Any:
        if data and isinstance(data, str):
            return [data]
        return data

    @field_validator("parent_dir", mode="before", json_schema_input_type=str)
    def __cast_str_to_path(cls, data: Any) -> Any:
        """Cast type of the `parent_dir` field. It will convert to Path object
        if it passes with string type.

        Args:
            data (Any): An any data that pass before validation.
        """
        return Path(data) if data and isinstance(data, str) else data

    @abstractmethod
    def execute(
        self,
        context: Context,
        engine: DictData,
        metric: MetricEngine,
    ) -> Any:
        """Abstract execute method for any engine should implement before
        plugin to this package.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricEngine): A metric engine that was set from handler
                step for passing custom metric data.

        Notes:
            It same some recommended rule that it should to follow:
            - Should to use `handle_{layer}` method like
                - `handle_load`
                - `handle_apply`
                - `handle_save`
              instead of use its method directly.
            - Execution should not create any context. That mean the handle
              will do this task.

        Returns:
            Any: An Any DataFrame API specific with implemented engine.
        """

    @abstractmethod
    def set_result(self, df: Any, context: Context) -> Result:
        """Abstract set a Result object before returning from the handle_execute
        method. The method should implement from sub-model of this `BaseEngine`.

        Args:
            df (Any): An Any DataFrame API.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config model.
        """

    @abstractmethod
    def apply(
        self,
        df: Any,
        context: Context,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> Any:
        """Abstract apply method for start running the engine transform for each
        operator.

        Args:
            df (Any):
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricTransform): A metric transform that was set from
                handler step for passing custom metric data.
        """

    def post_execute(
        self,
        context: Context,
        *,
        engine: DictData,
        exception: Exception | None = None,
        **kwargs,
    ) -> Any:
        """Overridable Post Execute method.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            exception (Exception | None, default is None) An exception object if
                the returning execute method raise some error while executing.

        Returns:
            Any: An any output that will update to the context with key
                `post_execution_output`.
        """

    @abstractmethod
    def set_engine_context(self, context: Context, **kwargs) -> DictData:
        """Set Engine Context data for passing to execute method.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.

        Returns:
            DictData: An any engine context data.
        """

    def handle_execute(self, context: Context) -> Result:
        """Handle execute method with preparing any DataFrame API to the Result
        object. This method must call from the Tool Operator execute method.

            This engine handler execute method will pre-define engine metric
        before call abstract execute method.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.

        Raises:
            Exception: An any exception that is able to catch from the `execute`
                or `set_result` methods.

        Returns:
            Result: A Result model that catch from the sink DataFrame API info.
        """
        name: str = self.app_name or self.name
        logger.info(f"👷🏗️ Handle Execute config name: {name!r} ...")
        metric = MetricEngine(app_id=name, type=self.type)
        context.update({"metric_engine": metric})
        engine: DictData = self.set_engine_context(context)
        exception: Exception | None = None
        try:
            rs = self.execute(context, engine=engine, metric=metric)
            return self.set_result(rs, context)
        except Exception as e:
            exception = e
            logger.exception(
                f"Catch error from the engine execution: {e.__class__.__name__}"
            )
            raise
        finally:
            logger.warning("Start Post execution and update metric extra ...")
            post_output: Any = self.post_execute(
                context, engine=engine, exception=exception
            )
            metric.finish()
            context.update(
                {
                    "metric_extras": engine.pop("metric_extras", {}),
                    "post_execution_output": post_output,
                }
            )

    def handle_apply(
        self,
        df: Any,
        context: Context,
        engine: DictData,
        **kwargs,
    ) -> Any:
        """Handle Apply Engine transform method. This method will update
        necessary transform metrics before and catch up it after apply.

            This handle method will update `metric_operator` and `metric_group_transform`
        keys to the context object before start its apply method. After apply,
        it will prepare the keys that set before to `metric_transform` key

        Args:
            df (Any): An any DataFrame API that use to apply transform.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.

        Returns:
            Any: Return an any DataFrameAPI that was applied with all transform
                operator.
        """
        metric = MetricTransform()
        metric_op_group = MetricOperatorGroup(type="group")
        context.update(
            {
                "metric_transform": metric,
                "metric_operator": [],
                "metric_group_transform": metric_op_group,
            }
        )
        try:
            df: Any = self.apply(
                df, context, engine=engine, metric=metric, **kwargs
            )
            return df
        except Exception as e:
            logger.error(
                f"Catch error from the engine apply: {e.__class__.__name__}"
            )
            raise
        finally:
            metric_op_group.finish()
            metric.transforms = context.get("metric_operator", [])
            metric.finish()

    def emit(self, context: Context) -> None:
        """Emit metric data to each metric model with the current context data.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
        """
        data = MetricData.model_validate(context)
        for metric in self.metrics:
            if metric.condition != ALWAYS and (
                (metric.condition == ONLY_FAILED and data.run_result)
                or (metric.condition == ONLY_SUCCESS and not data.run_result)
            ):
                continue
            try:
                logger.info(
                    f"👷🏗📩 Handler Emit with metric type: {metric.type!r}"
                )
                metric.handle_emit(data=data, **context)
            except Exception as e:
                logger.exception(
                    f"‼️ Error occurred when trying to push metric: "
                    f"{metric.type!r} with err: {e.__class__.__name__}, "
                    f"continue to next emitter\n... {e}"
                )


class BaseSource(BaseModel, ABC):
    """Base Source abstract model."""

    type: str = Field(description="A source type.")
    sample_records: int | None = Field(
        default=None,
        description="A sample records value that use to limit data after load.",
        ge=0,
        le=10_000,
    )

    @field_validator("sample_records")
    def check_range(cls, value: Any) -> Any:
        """Validate number of sample records."""
        if value is not None and (value < 0 or value > 10_000):
            raise ValueError(
                "sample record should has range between 0 and 10000"
            )
        return value

    @abstractmethod
    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[Any, Shape]:
        """Load abstract method for this source model.

        Args:
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricSource): A metric source that was set from handler
                step for passing custom metric data.
        """

    @abstractmethod
    def inlet(self) -> tuple[str, str]:
        """Return a pair of type and name of the inlet data for the event hub
        service for create data lineage together with the outlet of sink.

        Returns:
            tuple[str, str]: A pair of type and this source name.
        """

    def handle_load(
        self,
        context: Context,
        *,
        engine: DictData | None = None,
        **kwargs,
    ) -> Any:
        """Handle load method with adding source metric to the execution
        context.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
        """
        logger.info(f"👷🚰 Handle Source: {self.type!r} ...")
        metric = MetricSource(type=self.type, inlet=self.inlet())
        context.update({"metric_source": metric})
        try:
            df, shape = self.load(engine=engine, metric=metric, **kwargs)
            metric.read_row_count = shape.rows
            metric.read_column_count = shape.columns
            return df
        finally:
            metric.finish()


class BaseTransform(BaseModel, ABC):
    """Base Transform abstract model for any implemented transform operator
    for each engine. That mean the transform operator have different supported
    operator base on engine that it implemented for.

    AbstractMethods:
        apply:

    Methods:
        handle_apply:

    ClassMethods:
        get_op_support: Return a tuple of supported operator type name.
    """

    op: str = Field(description="An operator type.")

    @abstractmethod
    def apply(
        self,
        df: Any,
        engine: DictData,
        metric: MetricOperatorOrder,
        **kwargs,
    ) -> Any:
        """Apply operator transform abstraction method.

        Args:
            df (Any): An any DataFrame API instance.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.
        """

    def handle_apply(
        self,
        df: Any,
        context: DictData,
        engine: DictData,
        **kwargs,
    ) -> Any:
        """Handle Apply sub-transform logic.

        Args:
            df (Any): An any DataFrame API instance.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
        """
        logger.info(f"👷🔧 Handle Apply Operator: {self.op!r}")
        metric = MetricOperatorOrder(type="order", trans_op=self.op)
        context["metric_operator"].append(metric)
        try:
            return self.apply(df, engine=engine, **kwargs)
        finally:
            self.post_apply(engine=engine, **kwargs)
            metric.finish()

    @classmethod
    def get_op_support(cls) -> tuple[str, ...]:
        """Get the tuple of supported operator for this transform model.

        Raises:
            TypeError: If override type of this `op` field does not be Literal.

        Returns:
            tuple[str, ...]: A tuple of operator name that was supported for
                this transform model.
        """
        ant = cls.model_fields["op"].annotation
        if get_origin(ant) is Literal:
            return get_args(ant)
        raise TypeError(f"Does not support `op` override with type: {ant}")

    @staticmethod
    @abstractmethod
    def sync_schema(pre, post, metric, **kwargs) -> None: ...

    def post_apply(self, engine: DictData, **kwargs): ...


class BaseSink(BaseModel, ABC):
    """Base Sink abstract model."""

    type: str = Field(description="A sink type.")

    @abstractmethod
    def save(
        self,
        df: Any,
        engine: DictData,
        metric: MetricSink,
        **kwargs,
    ) -> tuple[Any, Shape]:
        """Save abstract method for this sink model.

        Args:
            df (Any): An Any DataFrame API.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricSink): A metric sink that was set from handler step
                for passing custom metric data.

        Returns:
            tuple[Any, Shape]: A pair of an any DataFrame API and its Shape
                model.
        """

    @abstractmethod
    def outlet(self) -> tuple[str, str]:
        """Return a pair of type and name of the outlet data for the event hub
        service for create data lineage together with the inlet of source.

        Returns:
            tuple[str, str]: A pair of type and this sink name.
        """

    @abstractmethod
    def dest(self) -> str:
        """Return destination value of this sink model."""

    def handle_save(
        self,
        df: Any,
        context: Context,
        engine: DictData,
        **kwargs,
    ) -> Any:
        """Handle save method with adding source metric to the execution
        context.

        Args:
            df (Any): An Any DataFrame API.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.

        Returns:
            Any: An Any DataFrame API specific with implemented engine.
        """
        logger.info(f"👷🎯 Handle Sink: {self.type!r} ...")
        metric = MetricSink(
            type=self.type, outlet=self.outlet(), destination=self.dest()
        )
        context.update({"metric_sink": metric})
        try:
            df, shape = self.save(df, engine=engine, metric=metric, **kwargs)
            metric.write_row_count = shape.rows
            metric.write_column_count = shape.columns
            if "mode" in self.model_fields:
                # noinspection Pydantic
                metric.mode = self.mode
            return df
        finally:
            metric.finish()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated, Any, Literal, Union, get_type_hints

from pydantic import Field
from pydantic.functional_validators import field_validator

from ....__types import DictData
from ....errors import ToolTransformError
from ....models import Context, MetricOperatorGroup, MetricOperatorOrder
from .__abc import BasePolarsTransform, is_pair_expr
from .functions import (
    ExplodeArrayColumn,
    FlattenAllExceptArray,
    RenameColumns,
    RenameSnakeCase,
    Sql,
)
from .functions import Expr as ExprTransform

if TYPE_CHECKING:
    from polars import DataFrame, Schema
    from polars import Expr as Column

    PairCol = tuple[Column, str]

logger = logging.getLogger("jett")

GroupTransform = Annotated[
    Union[ExprTransform,],
    Field(
        discriminator="op",
        description="A transform that allow to use in group operator.",
    ),
]


class Group(BasePolarsTransform):
    """Group Transform model."""

    op: Literal["group"]
    transforms: list[GroupTransform] = Field(
        description=(
            "A list of operator transform model that support for grouping "
            "apply that mean it can execute with parallelism."
        )
    )

    # ARCHIVE:
    # @field_validator("transforms", mode="after")
    # def __validate_supported_output_type(
    #     cls,
    #     data: list[GroupTransform],
    # ) -> list[GroupTransform]:
    #     for t in data:
    #         hints: dict[str, Any] = get_type_hints(t)
    #         if (
    #             'return' in hints
    #             and (rt := hints['return'])
    #             and (
    #                 isinstance(rt, DataFrame)
    #                 or (
    #                     isinstance(rt, list)
    #                     and any(isinstance(i, DataFrame) for i in rt)
    #                 )
    #             )
    #         ):
    #             raise ValueError(
    #                 f"Group operator transform does not support operator: {t}."
    #             )
    #     return data

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperatorGroup,
        **kwargs,
    ) -> DataFrame:
        """Apply to Group operator transform.

        Args:
            df (Any): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.

        Returns:

        """
        maps: dict[str, Column] = {}
        for t in self.transforms:
            metric_order = MetricOperatorOrder(type="order", trans_op=t.op)
            metric.transforms.append(metric_order)
            try:
                output: PairCol | list[PairCol] | Any = t.apply(
                    df, engine=engine, metric=metric_order
                )

                # VALIDATE: Check the result type that match on the group operator.
                if is_pair_expr(output):
                    rs: dict[str, Column] = {output[1]: output[0]}
                elif isinstance(output, list) and all(
                    is_pair_expr(i) for i in output
                ):
                    rs: dict[str, Column] = {p[1]: p[0] for p in output}
                else:
                    raise ToolTransformError(
                        "Group on Spark engine should return the apply"
                        "group result with a pair of Column and str only."
                    )
                maps.update(rs)
            finally:
                metric_order.finish()
        df: DataFrame = df.with_columns(maps)
        return df

    def handle_apply(
        self,
        df: DataFrame,
        context: Context,
        engine: DictData,
        **kwargs,
    ) -> DataFrame:
        """Handle the Group Operator Apply.

        Args:
            df (DataFrame): A Polars DataFrame.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.

        Returns:
            DataFrame: A applied Spark DataFrame.
        """
        logger.info(f"👷🔧 Handle Apply Operator: {self.op!r}")
        metric = MetricOperatorGroup(type="group", trans_op=self.op)
        context["metric_operator"].append(metric)
        try:
            pre_schema: Schema = df.schema
            df: DataFrame = self.apply(df, engine, metric=metric, **kwargs)
            self.sync_schema(pre_schema, df.schema, metric=metric)
            return df
        finally:
            metric.finish()


Transform = Annotated[
    Union[
        ExplodeArrayColumn,
        RenameSnakeCase,
        RenameColumns,
        Sql,
        ExprTransform,
        FlattenAllExceptArray,
    ],
    Field(discriminator="op"),
]



import json
import sys
from functools import partial
from pathlib import Path

import click
from pydantic import TypeAdapter

from jett.__about__ import __version__
from jett.engine import Engine
from jett.tools import ToolModel

st_bold = partial(click.style, bold=True)
st_bold_red = partial(click.style, bold=True, fg="red")
st_bold_green = partial(click.style, bold=True, fg="green")
echo = click.echo


@click.group()
def cli() -> None:
    """Main Jett Command Line Interface ."""


@cli.command("version")
def version() -> None:
    """Return the current version of jett package that already installed
    on your Python environment.
    """
    echo(__version__)
    sys.exit(0)


@cli.command("json-schema")
@click.option(
    "--output-file",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        path_type=Path,
    ),
    help="JSON Schema filepath that want to write after generate.",
)
def gen_json_schema(output_file: Path | None = None) -> None:  # pragma: no cov
    """Generate JSON Schema for validate Jett configuration YAML template.

    \f
    Args:
        output_file (Path, default None): An output filepath that want to use
            for writing JSON Schema data after generated.
    """
    json_schema = TypeAdapter(Engine).json_schema(by_alias=True)
    template_schema: dict[str, str] = {
        "$schema": "http://json-schema.org/draft/2020-12/schema",
        "$id": "https://example.com/yapt.schema.json",
        "title": "Jett",
        "description": "DE Tool Configuration JSON Schema",
        "version": __version__,
    }
    echo(
        st_bold_green(
            "Start generate or rewrite the current JSON schema file ..."
        )
    )
    out_file = output_file or Path("./json-schema.json")
    with open(out_file, mode="w") as f:
        json.dump(template_schema | json_schema, f, indent=2)


@cli.command("validate")
@click.option(
    "--file",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        path_type=Path,
    ),
    help="A Jett config file path.",
)
def validate_conf(file: Path):
    """Validate Jett config file."""
    try:
        tool = ToolModel.from_yaml(path=file)
        echo(st_bold_green(f"✅ Tool config ({file}) is valid !!"))
        echo(f"> Name: {tool.model.name}")
        echo(f"> Author: {tool.model.author or 'Anon'}")
        echo(f"> Engine Type: {tool.model.type}")
    except Exception as e:
        echo(st_bold_red(f"❌ Tool config ({file}) is invalid !!"))
        raise e


@cli.command("run")
@click.option(
    "--file",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        path_type=Path,
    ),
)
def run(file: Path):
    # NOTE: Start Validate first.
    validate_conf(file)
    sys.exit(0)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pydantic import Field
from pydantic.functional_validators import model_validator
from typing_extensions import Self

from ....__types import DictData
from ....errors import ToolTransformError
from ....models import MetricOperatorOrder
from ....utils import to_snake_case
from ...__abc import BaseEngine
from ..utils import col_path, extract_cols_without_array
from .__abc import BasePolarsTransform, ColMap

if TYPE_CHECKING:
    from polars import DataFrame
    from polars import Expr as Column

    PairCol = tuple[Column, str]

logger = logging.getLogger("jett")


class RenameSnakeCase(BasePolarsTransform):
    """Rename All columns to Snakecase operator transform model."""

    op: Literal["rename_snakecase"] = Field(
        description="An operator transform type.",
    )

    def apply(self, df: DataFrame, engine: DictData, **kwargs) -> DataFrame:
        """Apply to Rename Column with Snakecase to the Polars DataFrame.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
        """
        renames: dict[str, str] = {c: to_snake_case(c) for c in df.columns}
        logger.info(
            f"Start Rename All columns to Snakecase:\n"
            f"{json.dumps(renames, indent=1)}"
        )
        return df.rename(renames)


class RenameColumns(BasePolarsTransform):
    """Rename Columns Transform model."""

    op: Literal["rename"]
    columns: list[ColMap] = Field(description="A list of ColMap object.")
    allow_group_transform: bool = Field(default=False)

    def apply(self, df: DataFrame, engine: DictData, **kwargs) -> DataFrame:
        """Apply to Rename Column to the Polars DataFrame.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
        """
        logger.info(f"Rename columns statement:\n{self.columns}")
        return df.rename({col.source: col.name for col in self.columns})


class Expr(BasePolarsTransform):
    op: Literal["expr"]
    name: str
    query: str

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperatorOrder,
        **kwargs,
    ) -> PairCol:
        """Apply to Expr.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.

        Returns:
            PairCol: A pair of Expr object and its alias name.
        """
        import polars as pl

        return pl.sql_expr(self.query), self.name


class Sql(BasePolarsTransform):
    op: Literal["sql"]
    sql: str | None = Field(default=None, description="A SQL statement.")
    sql_file: str | None = Field(
        default=None,
        description="A SQL file that want to load.",
    )

    @model_validator(mode="after")
    def __check_sql_and_sql_file(self) -> Self:
        """Check the necessary field of SQL use pass only SQL statement nor
        SQL filepath.
        """
        if not self.sql and not self.sql_file:
            raise ValueError("SQL and SQL file should not be empty together.")
        elif self.sql and self.sql_file:
            logger.warning(
                "If pass SQL statement and SQL file location together, it will "
                "use SQL statement first."
            )
        return self

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperatorOrder,
        **kwargs,
    ) -> DataFrame:
        """Apply to SQL query.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.

        Returns:
            DataFrame: A Polars DataFrame that have applied SQL query via `sql`
                method.
        """
        logger.info("Start Prepare SQL statement ...")
        if self.sql:
            sql: str = self.sql
        else:
            _engine: BaseEngine = engine["engine"]
            sql_file: Path = _engine.parent_dir / self.sql_file
            try:
                sql: str = sql_file.read_text()
            except FileNotFoundError:
                raise ToolTransformError(
                    f"SQL file does not exists {sql_file.resolve()}"
                ) from None
        logger.info(f"SQL Statement:\n{sql}")
        logger.info(f"{df.schema}")
        return df.sql(sql)


class SelectColumns(BasePolarsTransform):
    """Select Columns Operator transform model."""

    op: Literal["select"]
    columns: list[str]
    allow_missing: bool = Field(
        default=False,
        description=(
            "A flag to allow missing column when compare from the current "
            "DataFrame."
        ),
    )

    def apply(self, df: DataFrame, engine: DictData, **kwargs) -> DataFrame:
        """Apply to Select Column.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.

        Returns:
            DataFrame: A column selected Spark DataFrame.
        """
        selection: list[str] = self.columns
        if self.allow_missing:
            selection: list[str] = [c for c in self.columns if c in df.columns]
        return df.select(*selection)


class ExplodeArrayColumn(BasePolarsTransform):
    """Explode Array Column Operation transform model."""

    op: Literal["explode_array"]
    explode_col: str
    is_explode_outer: bool = True
    is_return_position: bool = False
    position_prefix_name: str = "_index_pos"

    def apply(self, df: DataFrame, engine: DictData, **kwargs) -> DataFrame:
        """Apply to Explode Array column.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.

        Returns:
            DataFrame: A Polars DataFrame that have exploded column.
        """
        import polars as pl

        if not self.is_return_position:
            logger.info("Start Explode Array Column")
            if self.is_explode_outer:
                return df.with_columns(pl.col(self.explode_col).explode())
            return df.with_columns(
                pl.col(self.explode_col).list.drop_nulls().explode()
            )
        raise NotImplementedError(
            "Does not support for return position column."
        )


class FlattenAllExceptArray(BasePolarsTransform):
    """Flatten all Columns except Array datatype Operator transform model."""

    op: Literal["flatten_all_except_array"]

    def apply(self, df: DataFrame, engine: DictData, **kwargs) -> DataFrame:
        """Apply to Flatten all Columns except Array or List data type.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.

        Returns:
            DataFrame: A Polars DataFrame that already
        """
        selection: dict[str, Column] = {}
        for c in extract_cols_without_array(schema=df.schema):
            selection["_".join(c.split("."))] = col_path(c)

        logger.info("Start Flatten all columns except array")
        for k, v in selection.items():
            logger.info(f"> Target col: {k}, from: {v}")
        return df.with_columns(**selection).select(*selection.keys())



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pydantic import Field

from .....__types import DictData
from .....models import Shape
from ....__abc import BaseSource

if TYPE_CHECKING:
    from polars import DataFrame


class LocalCSVFile(BaseSource):
    """Local file system with CSV file format source model."""

    type: Literal["local"] = Field(description="A type of source.")
    file_format: Literal["csv"] = Field(description="A file format.")
    path: str
    delimiter: str = "|"
    header: bool = Field(default=True)
    sample_records: int | None = 200

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        """Load CSV file to DuckDB Relation object."""
        import polars as pl

        file_format: str = Path(self.path).suffix
        if file_format not in (".csv",):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not support."
            )
        df: DataFrame = pl.read_csv(
            source=self.path,
            separator=self.delimiter,
            has_header=self.header,
            sample_size=self.sample_records,
        )
        return df, Shape.from_tuple(df.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalJsonFile(BaseSource):
    """Local JSON file data source."""

    type: Literal["local"]
    file_format: Literal["json", "ndjson"]
    path: str

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        """Load JSON file from local file system."""
        import polars as pl

        file_format: str = Path(self.path).suffix
        if file_format not in (".json",):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not support."
            )

        if self.file_format == "ndjson":
            df: DataFrame = pl.read_ndjson(source=self.path)
        else:
            df: DataFrame = pl.read_json(source=self.path)
        return df, Shape.from_tuple(df.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path



import os
from datetime import datetime
from pathlib import Path

from airflow.models import DAG
from airflow.operators.empty import EmptyOperator

from jett.plugins.airflow.operators import JettOperator

with DAG(
    dag_id="demo_jett",
    start_date=datetime(2025, 1, 1),
    schedule=None,
    catchup=False,
    template_searchpath=str((Path(__file__).parent / "assets").absolute()),
    user_defined_macros={"env": os.getenv},
) as dag:
    start = EmptyOperator(task_id="start", dag=dag)
    jett = JettOperator(task_id="jett", tool="arrow.csv.tool", dag=dag)
    start >> jett



from __future__ import annotations

import copy
import logging
import re
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.connect.session import SparkSession as SparkRemoteSession
    from pyspark.sql.types import DataType, StructField, StructType

from ...utils import handle_command, is_snake_case

logger = logging.getLogger("jett")

ENUM_EXTRACT_ARRAY_TYPE: Final[str] = "__array_element"
ALLOW_VALIDATE_PATTERNS: Final[tuple[str, ...]] = (
    "non_snake_case",
    "whitespace",
)


def schema2dict(
    schema: StructType, sorted_by_name: bool = False
) -> list[dict[str, str]]:
    """Convert StructType object to dict that include keys, `name` and `dtype`.

    Args:
        schema:
        sorted_by_name:
    """
    rs: list[dict[str, str]] = [
        {"name": c.name, "dtype": c.dataType.simpleString()} for c in schema
    ]
    return sorted(rs, key=lambda x: x["name"]) if sorted_by_name else rs


def is_remote_session(spark: SparkSession | SparkRemoteSession) -> bool:
    """Check whether given session is remote session or not."""
    from pyspark.sql.connect.session import SparkSession as SparkRemoteSession

    return isinstance(spark, SparkRemoteSession)


def yarn_kill(app_id: str) -> None:
    """Kill spark application on YARN."""
    kill_cmd: str = f"yarn application -kill {app_id}"
    for line in handle_command(cmd=kill_cmd):
        logger.info(line.strip())


def yarn_fetch_log(application_id: str, log_file: str) -> None:
    """Fetch the YARN application log (stdout and stderr)."""
    fetch_cmd: str = (
        f"yarn logs -applicationId {application_id} -log_files {log_file}"
    )
    for line in handle_command(cmd=fetch_cmd):
        logger.info(line.strip())


def extract_spark_conf_keys(cmd: list[str]) -> list[str]:
    """Extract spark conf keys from existing list of submit commands."""
    return [
        c.split("=")[0].replace("'", "").replace('"', "")
        for c in cmd
        if "=" in c and len(c) > 1 and c.count("=") == 1
    ]


def clean_tz_for_extra_java_options(option: str, tz: str) -> str:
    """Clean and set time zone in extraJavaOptions"""
    java_user_timezone_key: str = "user.timezone"
    options: list[str] = [
        p.strip()
        for p in option.split("-D")
        if java_user_timezone_key not in p.split("=")[0]
    ]
    options = list(filter(None, options))
    options.append(f"{java_user_timezone_key}={tz}")
    options = [f"-D{p}" for p in options]
    return " ".join(options)


def add_spark_cmd(
    cmd: list[str], key: str, value: str | int | list[str] | None
) -> list[str]:
    """Extend Spark command that generate after checking value is not empty."""
    if value is None:
        return cmd

    if isinstance(value, list) and len(value) != 0:
        value_str: str = ",".join(value)
        cmd.extend([f"--{key}", f'"{value_str}"'])
    elif isinstance(value, str):
        cmd.extend([f"--{key}", value])
    elif isinstance(value, int):
        cmd.extend([f"--{key}", str(value)])

    return cmd


def has_fix_array_index(input_str: str) -> bool:
    """Check that string contains bracket, `[{number}]`, or not

    Examples:
        contains case: items[0], items[322]
    """
    return re.search(r"\[\d+]", input_str) is not None


def replace_fix_array_index_with_x_index(input_str: str) -> str:
    """Replace any existing string like `[{number}]` with [x]

    Examples:
        input:
            >>> "items[0].detail[322].name"
        output:
            >>> "items[x].detail[x].name"
    """
    return re.sub(r"\[\d+]", "[x]", input_str)


def extract_cols_selectable(schema: StructType, prefix: str = "") -> list[str]:
    """Extracts all selectable columns of given schema, support all top level
    column and nested column and array column.

    Returns:
        list[str]: All cols like:
            ["c1", "c2.f1", "c2.f2", "c3"]

    Examples:
        Input:
        >>> from pyspark.sql.types import StringType, IntegerType, DoubleType
        >>> StructType(
        ...     [
        ...         StructField("texts", ArrayType(StringType())),
        ...         StructField("items", ArrayType(StructType(
        ...             [
        ...                 StructField("name", StringType()),
        ...                 StructField("price", IntegerType()),
        ...                 StructField("detail", ArrayType(StructType(
        ...                     [
        ...                         StructField("field1", StringType()),
        ...                         StructField("field2", DoubleType()),
        ...                     ]
        ...                 )))
        ...             ]
        ...         )))
        ...     ]
        ... )

        Output:
        >>> [
        ...     'texts',
        ...     'texts[x]',
        ...     'items',
        ...     'items[x]',
        ...     'items[x].name',
        ...     'items[x].price',
        ...     'items[x].detail',
        ...     'items[x].detail[x]',
        ...     'items[x].detail[x].field1',
        ...     'items[x].detail[x].field2',
        ... ]
    """
    from pyspark.sql.types import ArrayType, StructType

    rs: list[str] = []
    for field in schema:
        rs.append(prefix + field.name)
        field_type = field.dataType
        if isinstance(field_type, StructType):
            rs.extend(
                extract_cols_selectable(field_type, f"{prefix}{field.name}.")
            )
        elif isinstance(field_type, ArrayType):
            rs.append(prefix + field.name + "[x]")
            if isinstance(field_type.elementType, StructType):
                rs.extend(
                    extract_cols_selectable(
                        field_type.elementType, f"{prefix}{field.name}[x]."
                    )
                )
    return rs


def replace_all_occurrences(input_str: str, old: str, new: str) -> str:
    """Replace all occurrences of a specific string (case-insensitive) with a
    given string.
    """
    return re.sub(re.escape(old), new, input_str, flags=re.IGNORECASE)


def is_table_exist(
    spark: SparkSession | SparkRemoteSession,
    database: str,
    table_name: str,
) -> bool:
    """Check table's existence."""
    catalog = spark.catalog.currentCatalog()
    return spark.catalog.tableExists(f"{catalog}.{database}.{table_name}")


def is_database_exist(
    spark: SparkSession | SparkRemoteSession, database: str
) -> bool:
    """Check database's existence."""
    return spark.catalog.databaseExists(dbName=database)


def extract_cols_without_array(schema: StructType) -> list[str]:
    """Extract selectable columns without array type.

        It returns only list of selectable columns that are not nested array
    type return only root array column name.

    Args:
        schema (StructType):

    Returns:
        list[str]: A list of column name that extract by selectable without
            array type.
    """
    selectable_cols: list[str] = extract_cols_selectable(schema=schema)
    nested_array_cols: list[str] = [c for c in selectable_cols if "[x]" in c]
    final_selectable_cols: list[str] = [
        c for c in selectable_cols if c not in nested_array_cols
    ]

    rs: list[str] = []
    for c in final_selectable_cols:
        not_parent: bool = True
        for fc in final_selectable_cols:
            if fc != c and fc.startswith(f"{c}."):
                not_parent: bool = False

        if not_parent:
            rs.append(c)
    return rs


def extract_col_with_pattern(
    schema: StructType,
    patterns: list[str],
    parent_cols: list[str] | None = None,
) -> list[str]:
    """Do recursive find the colum name that does not follow the pattern
    current supported patterns are non_snake_case and whitespace
    """
    from pyspark.sql.types import ArrayType, StructType

    parent_cols: list[str] = parent_cols or []

    def _validate_and_append_error(
        col_name: str,
        p_cols: list[str],
        e_cols: list[str],
        pt: list[str],
    ) -> list[str]:
        """Child Wrapped function for extract_pyspark_column_names_with_pattern
        validate snake case or whitespace and append error columns

        Args:
            col_name:
            p_cols: A parent columns
            e_cols: An error columns
            pt:
        """
        is_found_err: bool = False
        for pattern in pt:
            if pattern == "non_snake_case" and not is_snake_case(col_name):
                is_found_err = True
            elif pattern == "whitespace" and " " in col_name:
                is_found_err = True

        if is_found_err:
            error_col = (
                col_name
                if len(p_cols) == 0
                else ".".join(p_cols) + f".{col_name}"
            )
            e_cols.append(error_col)

        return e_cols

    if all(p not in ALLOW_VALIDATE_PATTERNS for p in patterns):
        raise ValueError(
            f"patterns must contain value in {ALLOW_VALIDATE_PATTERNS}"
        )

    error_cols: list[str] = []

    for column in schema:
        if isinstance(column.dataType, ArrayType):
            if isinstance(column.dataType.elementType, StructType):
                _parent_cols = copy.deepcopy(parent_cols)
                _parent_cols.append(column.name)
                error_cols = error_cols + extract_col_with_pattern(
                    schema=column.dataType.elementType,
                    patterns=patterns,
                    parent_cols=_parent_cols,
                )
            else:
                error_cols = _validate_and_append_error(
                    col_name=column.name,
                    p_cols=parent_cols,
                    e_cols=error_cols,
                    pt=patterns,
                )
        elif isinstance(column.dataType, StructType):
            _parent_cols = copy.deepcopy(parent_cols)
            _parent_cols.append(column.name)
            error_cols = error_cols + extract_col_with_pattern(
                schema=column.dataType,
                patterns=patterns,
                parent_cols=_parent_cols,
            )
        else:
            error_cols = _validate_and_append_error(
                col_name=column.name,
                p_cols=parent_cols,
                e_cols=error_cols,
                pt=patterns,
            )
    return error_cols


def validate_col_disallow_pattern(
    schema: StructType, patterns: list[str]
) -> None:
    """Validate columns names in dataframe (support nested schema) from the
    pattern.
    """
    if all(p not in ALLOW_VALIDATE_PATTERNS for p in patterns):
        raise ValueError(
            f"Patterns must contain value in {ALLOW_VALIDATE_PATTERNS}"
        )

    error_cols = extract_col_with_pattern(schema=schema, patterns=patterns)
    if len(error_cols) > 0:
        cols: str = ", ".join(error_cols)
        raise ValueError(
            f"Please check column naming convention (must not be {patterns!r}) "
            f"on columns: {cols}"
        )


def validate_col_allow_snake_case(schema: StructType) -> None:
    """Validate columns, allow only snake case."""
    validate_col_disallow_pattern(schema=schema, patterns=["non_snake_case"])


def get_simple_str_dtype(dtype: DataType | StructField) -> str:
    """Get pyspark data type in simple string format."""
    from pyspark.sql.types import StructField

    return (
        dtype.dataType.simpleString()
        if isinstance(dtype, StructField)
        else dtype.simpleString()
    )


def is_root_level_column(schema: StructType, column: str) -> bool:
    """Return True if it is a column root level column."""
    for schema_col in schema:
        if schema_col.name == column:
            return True
    return False


def is_root_level_column_primitive_type(
    schema: StructType, column: str
) -> bool:
    """is a column is a root level column and also a primitive type."""
    from pyspark.sql.types import ArrayType, StructType

    is_root_column = is_root_level_column(schema=schema, column=column)
    if not is_root_column:
        return is_root_column

    for schema_col in schema:
        if schema_col.name == column and not isinstance(
            schema_col.dataType, (ArrayType, StructType)
        ):
            return True
    return False


def _extract_selectable_plan_with_auto_explode(
    schema: DataType,
    columns: list[str],
    select_columns: list[str] | None = None,
    plans: list[dict] | None = None,
) -> list[dict]:
    """A child function for method `select_with_auto_explode`
    extract the plan of select and explode based on the given schema.

    Args:
        schema (DataType):
        columns (list[str]):
        select_columns:
        plans:
    """
    from pyspark.sql.types import ArrayType, StructType

    select_columns: list[str] = select_columns or []
    plans: list[dict] = plans or []

    if len(columns) == 0:
        if len(select_columns) > 0:
            plans.append(
                {"action": "select", "columns": ".".join(select_columns)}
            )
        return plans

    if not isinstance(schema, StructType):
        raise ValueError("schema mismatch, schema must be StructType")

    current_column = columns.pop(0)
    current_field = [s for s in schema if s.name == current_column]
    if len(current_field) == 0:
        raise ValueError(
            f"not found column {current_column} in current struct {schema.simpleString()}"
        )

    current_field = current_field[0]
    child_field = current_field.dataType
    if isinstance(current_field.dataType, ArrayType) and len(columns) > 0:
        if len(select_columns) > 0:
            select_columns.append(current_column)
            plans.append(
                {"action": "select", "columns": ".".join(select_columns)}
            )
            select_columns = []

        if len(columns) > 0:
            plans.append(
                {
                    "action": "explode",
                    "columns": current_column,
                    "alias": "temp_array_col",
                }
            )
            select_columns.append("temp_array_col")

        child_field = current_field.dataType.elementType
    else:
        select_columns.append(current_column)

    return _extract_selectable_plan_with_auto_explode(
        schema=child_field,
        columns=columns,
        select_columns=select_columns,
        plans=plans,
    )


def select_with_auto_explode(df: DataFrame, column: str) -> DataFrame:
    """This method does similar thing to df.select(), but it explodes array
    column automatically when the schema during column extraction is ArrayType.
    """
    from pyspark.sql.functions import explode

    plans = _extract_selectable_plan_with_auto_explode(
        schema=df.schema, columns=column.split("."), select_columns=[], plans=[]
    )
    rs_df: DataFrame = df
    for i, plan in enumerate(plans):
        _col = plan["columns"]
        if plan["action"] == "explode":
            _col = explode(_col).alias(plan["alias"])

        if i == 0:
            rs_df = df.select(_col)
        else:
            rs_df = rs_df.select(_col)
    return rs_df


def column_has_only_null_or_empty_array(df: DataFrame, column: str) -> bool:
    """
    check that the column contains only null or empty array
    cannot handle case array -> struct -> array
    """
    new_df = select_with_auto_explode(df=df, column=column)
    rows = new_df.distinct().collect()
    has_only_null_or_empty_array = True
    keywords = [None, [None], []]
    for row in rows:
        for _, v in row.asDict(True).items():
            if v not in keywords:
                has_only_null_or_empty_array = False

    return has_only_null_or_empty_array


def _child_remove_column_from_struct(
    schema_dict: dict, columns: list[str]
) -> dict:
    """A child function of remove_column_from_struct recursively extract the
    dict of pyspark schema and remove the given column.
    """
    # NOTE: remove the column
    if len(columns) == 1:
        if schema_dict["type"] != "struct":
            raise ValueError(
                f"schema must be struct, cannot delete the column {columns}"
            )

        new_fields = [
            field
            for field in schema_dict["fields"]
            if field["name"] != columns[0]
        ]
        if len(new_fields) == len(schema_dict["fields"]):
            raise ValueError(
                f"field {columns[0]} not found in current struct: {schema_dict}"
            )

        schema_dict["fields"] = new_fields
        return schema_dict

    current_column = columns.pop(0)
    index = None
    current_field = None

    if schema_dict["type"] == "struct":
        for i, field in enumerate(schema_dict["fields"]):
            if field["name"] == current_column:
                current_field = field
                index = i
                break
        if index is None:
            raise ValueError(
                f"field {current_column} not found in current schema "
                f"{schema_dict}"
            )

        if (
            current_field["type"]["type"] == "array"
            and len(columns) != 0
            and columns[0] != ENUM_EXTRACT_ARRAY_TYPE
        ):
            raise ValueError(
                f"schema mismatch, found array schema but parent field name "
                f"{columns[0]} is not {ENUM_EXTRACT_ARRAY_TYPE}"
            )

        schema_dict["fields"][index]["type"] = _child_remove_column_from_struct(
            schema_dict=current_field["type"], columns=columns
        )
    elif schema_dict["type"] == "array":
        schema_dict["elementType"] = _child_remove_column_from_struct(
            schema_dict=schema_dict["elementType"], columns=columns
        )

    return schema_dict


def remove_column_from_struct(schema: StructType, column: str) -> StructType:
    """Remove column from given schema (StructType)."""
    from pyspark.sql.types import StructType

    schema_dict = _child_remove_column_from_struct(
        schema_dict=schema.jsonValue(),
        columns=column.split("."),
    )
    return StructType.fromJson(json=schema_dict)


def _child_replace_column_dtype_in_nested_schema(
    schema_dict: dict, parent_cols: list[str], target_column: str, to_dtype: str
) -> dict:
    """A child function of replace_column_dtype_in_nested_schema recursively
    manipulate pyspark schema in JSON format replace data type from given param
    """

    def __get_index_by_field_name(schema: dict, field_name: str) -> int:
        fields = schema["fields"]
        idx = [
            i for i, field in enumerate(fields) if field["name"] == field_name
        ]
        if len(idx) == 0:
            raise ValueError(
                f"not found field {field_name} (parent col) in struct {fields}"
            )
        return idx[0]

    def __classify_schema_type(schema: dict) -> str:
        _schema_type = None
        current_schema_type = schema["type"]
        if isinstance(current_schema_type, dict):
            return "root_struct"
        elif current_schema_type == "struct":
            return "nested_struct"
        elif current_schema_type == "array":
            if isinstance(schema.get("elementType"), str):
                _schema_type = "array_primitive"
            else:
                _schema_type = "array_struct"
        return _schema_type

    if len(parent_cols) == 0:
        index = __get_index_by_field_name(
            schema=schema_dict, field_name=target_column
        )
        element = schema_dict["fields"][index]
        element["type"] = to_dtype
        schema_dict["fields"][index] = element
        return schema_dict

    first_parent_col = parent_cols.pop(0)
    schema_type = __classify_schema_type(schema=schema_dict)

    # handle case's root schema
    if schema_type == "root_struct" and schema_dict["name"] == first_parent_col:
        schema_dict["type"] = _child_replace_column_dtype_in_nested_schema(
            schema_dict=schema_dict["type"],
            parent_cols=parent_cols,
            target_column=target_column,
            to_dtype=to_dtype,
        )
        return schema_dict

    if schema_type == "array_primitive":
        schema_dict["elementType"] = to_dtype
        return schema_dict

    if schema_type == "array_struct":
        schema_dict["elementType"] = (
            _child_replace_column_dtype_in_nested_schema(
                schema_dict=schema_dict["elementType"],
                parent_cols=parent_cols,
                target_column=target_column,
                to_dtype=to_dtype,
            )
        )
        return schema_dict

    # get next struct field
    index = __get_index_by_field_name(
        schema=schema_dict, field_name=first_parent_col
    )
    schema_dict["fields"][index]["type"] = (
        _child_replace_column_dtype_in_nested_schema(
            schema_dict=schema_dict["fields"][index]["type"],
            parent_cols=parent_cols,
            target_column=target_column,
            to_dtype=to_dtype,
        )
    )
    return schema_dict


def replace_column_dtype_in_nested_schema(
    schema: StructField,
    parent_cols: list[str],
    target_column: str,
    to_dtype: DataType,
) -> StructField:
    """Replace target column to desire data type (support nested schema)."""
    from pyspark.sql.types import ArrayType, StructField, StructType

    if not isinstance(schema, StructField) or not isinstance(
        schema.dataType, (StructType, ArrayType)
    ):
        raise ValueError(
            "schema must be a struct of struct or struct of array only"
        )

    to_dtype = to_dtype.simpleString()
    if to_dtype == "bigint":
        to_dtype = "long"
    output_dict = _child_replace_column_dtype_in_nested_schema(
        schema_dict=schema.jsonValue(),
        parent_cols=parent_cols,
        target_column=target_column,
        to_dtype=to_dtype,
    )
    new_schema = StructField.fromJson(output_dict)
    return new_schema


def is_column_has_null(df: DataFrame, column: str) -> bool:
    """Check that the given column contains null value or not."""
    from pyspark.sql.functions import col

    return bool(df.filter(col(column).isNull()).first())


def is_timestamp_column(column_name: str, schema: StructType) -> bool:
    """Return true if the given column a timestamp column."""
    from pyspark.sql.types import TimestampNTZType, TimestampType

    is_timestamp_col = False
    for column in schema:
        if column.name == column_name and isinstance(
            column.dataType, (TimestampNTZType, TimestampType)
        ):
            is_timestamp_col = True
    return is_timestamp_col


def get_first_level_column_exclude_nested_struct(df: DataFrame) -> list[str]:
    """Get a list of first column level that does not have nested struct.

    Args:
        df (DataFrame): A Spark DataFrame.
    """
    columns: list[str] = []
    for c in df.schema:
        _s = c.simpleString().split(":")
        if "struct" not in _s[1]:
            columns.append(_s[0])
    return columns



from pydantic import BaseModel, Field


class ColumnMap(BaseModel):
    """Column Map model."""

    name: str = Field(description="A new column name.")
    source: str = Field(
        description="A source column statement before alias with alias.",
    )
    allow_quote: bool = True

    def gen_source(self) -> str:
        """Generate source statement with quote flag."""
        return (
            f"{self.source} AS {self.name}"
            if self.allow_quote
            else f'"{self.source}" AS {self.name}'
        )



import logging
from typing import Literal

import polars as pl
from polars import DataFrame
from pydantic import Field
from pyiceberg.types import (
    BinaryType,
    BooleanType,
    DateType,
    DecimalType,
    DoubleType,
    # DurationType,
    FloatType,
    IntegerType,
    ListType,
    LongType,
    # NullType,
    StringType,
    StructType,
    TimestampType,
    TimeType,
)

from ....__types import DictData
from ....models import MetricSink, Shape
from ...__abc import BaseSink
from ..utils import validate_col_allow_snake_case

logger = logging.getLogger("jett")

iceberg_type_mapping = {
    pl.Boolean: BooleanType,
    pl.Int8: IntegerType,
    pl.Int16: IntegerType,
    pl.Int32: IntegerType,
    pl.Int64: LongType,
    pl.UInt8: IntegerType,
    pl.UInt16: IntegerType,
    pl.UInt32: IntegerType,
    pl.UInt64: LongType,
    pl.Float32: FloatType,
    pl.Float64: DoubleType,
    pl.Decimal: DecimalType,
    pl.Utf8: StringType,
    pl.Binary: BinaryType,
    pl.Date: DateType,
    pl.Time: TimeType,
    pl.Datetime: TimestampType,
    # pl.Duration: DurationType,
    pl.List: ListType,
    pl.Categorical: StringType,
    pl.Struct: StructType,
    pl.Object: None,  # Not directly supported
    # pl.Null: NullType,
}


class Iceberg(BaseSink):
    type: Literal["iceberg"]
    mode: Literal["overwrite", "append", "merge"] = Field(
        default="overwrite",
        description=(
            "A write mode that should be `append`, `overwrite`, or `merge` "
            "only."
        ),
    )
    database: str
    table_name: str
    validate_column_snake_case: bool = True

    def save(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricSink,
        **kwargs,
    ) -> tuple[DataFrame, Shape]:
        """Save Polars DataFrame to the Iceberg table.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricSink): A metric sink that was set from handler step
                for passing custom metric data.

        Returns:
            tuple[DataFrame, Shape]:
        """
        logger.info("Sink - Start sink to Iceberg.")
        if df.is_empty():
            logger.warning("Sink - Found empty DataFrame, Skip writing!!!")
            metric.add("affected_partitions", 0)
            return df, Shape(rows=0, columns=0)

        if self.validate_column_snake_case:
            validate_col_allow_snake_case(schema=df.schema)

        return df, Shape()

    def outlet(self) -> tuple[str, str]:
        return "iceberg", self.dest()

    def dest(self) -> str:
        return self.table_name



from __future__ import annotations

import logging
from abc import ABC
from typing import TYPE_CHECKING, Literal

from pydantic import Field

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.connect.session import SparkSession as SparkRemoteSession

    AnySparkSession = SparkSession | SparkRemoteSession

from .....__types import DictData
from .....models import BasicFilter, Shape
from .....utils import bool2str
from ....__abc import BaseSource
from ...utils import is_remote_session

logger = logging.getLogger("jett")


class BaseGCSFile(BaseSource, ABC):
    type: Literal["gcs"]
    path: str
    file_format: str
    auth_type: str = "SERVICE_ACCOUNT_JSON_KEYFILE"
    auth_service_account_json_keyfile: str | None = None
    impl: str = "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem"
    sample_records: int | None = None
    filter: list[BasicFilter] = Field(default_factory=list)

    def set_gcs_connection(self, spark: AnySparkSession) -> None:
        """Set gcs connection in SparkContext."""
        logger.info("set fs.gs.impl=%s", self.impl)
        logger.info("set fs.gs.auth.type=%s", self.auth_type)
        logger.info(
            "set fs.gs.auth.service.account.json.keyfile=%s",
            self.auth_service_account_json_keyfile,
        )

        hadoop_conf = spark.sparkContext._jsc.hadoopConfiguration()
        hadoop_conf.set("fs.gs.impl", self.impl)
        hadoop_conf.set("fs.gs.auth.type", self.auth_type)

        # TODO
        # need to refactor this flow if we gonna support more auth type
        hadoop_conf.set(
            "fs.gs.auth.service.account.json.keyfile",
            self.auth_service_account_json_keyfile,
        )

    def inlet(self) -> tuple[str, str]:
        return "gcs", self.path


class GCSCSVFile(BaseGCSFile):
    file_format: Literal["csv"]
    delimiter: str = "|"
    header: bool = Field(default=True)

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        """Apply GCS CSV Loading process with the Spark session object."""
        spark: SparkSession = engine["spark"]
        if is_remote_session(spark):
            raise RuntimeError(
                f"Source: {self.type} does not support Spark Connect Session"
            )
        logger.info(f"🚰 Source - Start Load GCS: {self.path}")
        self.set_gcs_connection(spark=spark)
        reader = spark.read.format(self.file_format)

        df: DataFrame = (
            reader.option("header", bool2str(self.header))
            .option("delimiter", self.delimiter)
            .option("inferSchema", "true")
            .load(self.path)
        )

        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"🚰 Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        rows: int = df.count()
        shape: tuple[int, int] = (rows, len(df.columns))
        return df, Shape.from_tuple(shape)


class GCSJsonFile(BaseGCSFile):
    file_format: Literal["json"]
    multiline: bool = False

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        spark: SparkSession = engine["spark"]
        if is_remote_session(spark):
            raise RuntimeError(
                f"Source: {self.type} does not support Spark Connect Session"
            )
        logger.info(f"🚰 Source - Start Load GCS: {self.path}")
        self.set_gcs_connection(spark=spark)
        reader = spark.read.format(self.file_format)

        df: DataFrame = (
            reader.option("inferSchema", "true")
            .option("multiline", bool2str(self.multiline))
            .load(self.path)
        )

        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"🚰 Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        rows: int = df.count()
        shape: tuple[int, int] = (rows, len(df.columns))
        return df, Shape.from_tuple(shape)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal

from ....__types import DictData
from ....models import Shape
from ...__abc import BaseSink

if TYPE_CHECKING:
    from polars import DataFrame

logger = logging.getLogger("jett")


class Console(BaseSink):
    """Console Polars Sink model."""

    type: Literal["console"]
    limit: int = 10
    max_width: int | None = None

    def save(
        self,
        df: DataFrame,
        *,
        engine: DictData,
        **kwargs,
    ) -> Any:
        """Save the result data to the Console."""
        import polars as pl

        logger.info("🎯 Sink - Start sync with console")
        with pl.Config(set_tbl_width_chars=self.max_width):
            print(df.head(n=self.limit))
        return df, Shape.from_tuple(df.shape)

    def outlet(self) -> tuple[str, str]:
        return "console", self.dest()

    def dest(self) -> str:
        return "console"



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pydantic import Field

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession

from ....__types import DictData
from ..utils import validate_col_disallow_pattern
from .__abc import BaseSparkTransform


class ValidateColumnDisallowSpace(BaseSparkTransform):
    op: Literal["validate_col_names_disallow_whitespace"] = Field(
        description="An operator transform type."
    )

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply to Validate Column name disallow whitespace.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            DataFrame:
        """
        validate_col_disallow_pattern(schema=df.schema, patterns=["whitespace"])
        return df



from __future__ import annotations

import logging
import os
import shutil
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.connect.session import SparkSession as SparkRemoteSession
    from pyspark.sql.types import StructType

    AnySparkSession = SparkSession | SparkRemoteSession

from jett.utils import get_random_str, spark_env

from .utils import is_remote_session

logger = logging.getLogger("jett")


class TempStorage:
    """Temp Storage class for managing the DataFrame to temporary storage."""

    def __init__(
        self,
        spark: AnySparkSession,
        *,
        prefix: str | None = None,
    ) -> None:
        self.spark = spark
        self.temp_dir = self.get_temp_dir(prefix or "")
        self.is_written: bool = False

    def get_temp_dir(self, prefix: str) -> str:
        """Get the random string and set the temporary directory.

        Args:
            prefix (str): A prefix of dir.
        """
        dir_name: str = f"{prefix}" if prefix != "" else ""
        if is_remote_session(self.spark):
            dir_name = f"spark__connect__{dir_name}"

        if dir_name != "" and not dir_name.endswith("__"):
            dir_name = f"{dir_name}__"
        return os.path.join(
            spark_env("SPARK_TEMP_DIR"),
            f"{dir_name}{get_random_str(n=20)}",
        )

    def apply(
        self,
        df: DataFrame,
        override_schema: StructType = None,
    ) -> DataFrame:
        """Write data to temporary directory and load the data from the
        temporary directory.
        """
        logger.info(f"Write DataFrame to temporary directory: {self.temp_dir}")
        df.write.format("parquet").mode("overwrite").save(self.temp_dir)

        logger.info(f"Load DataFrame from: {self.temp_dir}")
        if override_schema is not None:
            df = self.spark.read.schema(override_schema).parquet(self.temp_dir)
        else:
            df = self.spark.read.parquet(self.temp_dir)
        self.is_written = True
        return df

    def teardown(self) -> None:
        """Clean the temporary directory."""
        if self.temp_dir is None or not self.is_written:
            return

        logger.info(f"Clean temporary directory: {self.temp_dir}")
        env: str = spark_env("ENV", default="")
        if env == "aws":
            self.__overwrite_temp_dir_with_empty()
        elif env == "iu":
            if is_remote_session(spark=self.spark):
                self.__overwrite_temp_dir_with_empty()
            else:
                self.__clean_temp_dir_via_spark_context()
        elif env == "local":
            if os.path.exists(self.temp_dir):
                shutil.rmtree(self.temp_dir)
            logger.info(f"⚠️ Directory {self.temp_dir} is deleted")
        else:
            raise RuntimeError(
                f"not implemented for spark env:{env}",
            )

    def __overwrite_temp_dir_with_empty(self) -> None:
        """Clean temp directory on the target dir via overwrite empty DataFrame.

        Warnings:
            This is just a temporary solution for any non-support valid
        `clean_temp_dir` process, this method will only overwrite temp_dir with
        empty dataframe for reducing usage of storage.
        """
        from pyspark.sql.types import StringType, StructField, StructType

        logger.info(
            "overwrite directory %s with empty dataframe data file",
            self.temp_dir,
        )
        empty_df = self.spark.createDataFrame(
            [], schema=StructType([StructField("dummy_col", StringType())])
        )
        empty_df.write.format("parquet").mode("overwrite").save(self.temp_dir)

    def __clean_temp_dir_via_spark_context(self) -> None:
        """Clean temp directory on the HDFS via Spark context."""
        sc = self.spark.sparkContext
        # ignore pylint protected-access
        fs = sc._jvm.org.apache.hadoop.fs.FileSystem.get(
            sc._jsc.hadoopConfiguration()
        )
        if fs.exists(sc._jvm.org.apache.hadoop.fs.Path(self.temp_dir)):
            fs.delete(sc._jvm.org.apache.hadoop.fs.Path(self.temp_dir))
            logger.info(f"⚠️ Directory {self.temp_dir} is deleted")



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame, SparkSession
    from pyspark.sql.connect.session import DataFrame as DataFrameRemote

    from ..sink import Sink

    PairCol = tuple[Column, str]
    AnyDataFrame = DataFrame | DataFrameRemote
    AnyApplyGroupOutput = PairCol | list[PairCol]
    AnyApplyOutput = AnyApplyGroupOutput | AnyDataFrame

from pydantic import Field

from ....__types import DictData
from ....errors import ToolTransformError
from ....utils import dt2str
from ..schema_change import (
    ENUM_EXTRACT_ARRAY_TYPE,
    evaluate_schema_change,
)
from ..utils import (
    get_simple_str_dtype,
    is_root_level_column_primitive_type,
    is_table_exist,
)
from .__abc import BaseSparkTransform

logger = logging.getLogger("jett")


class CalculateMinMaxOfColumns(BaseSparkTransform):
    """Calculate Min Max of Columns operator transform model."""

    op: Literal["calculate_min_max_of_columns"]
    columns: list[str] = Field()

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyApplyOutput:
        """Apply to Calculate the minimum and maximum values of given columns
        only support root level column and primitive type and result is stored
        in metric transform.
        """
        exception_group: list[str] = []
        for column in self.columns:
            is_safe = is_root_level_column_primitive_type(
                schema=df.schema, column=column
            )
            if not is_safe:
                exception_group.append(
                    f"{column} is not root level column and not primitive type"
                )

        if len(exception_group) != 0:
            raise ToolTransformError(
                f"failed to validate properties: {exception_group}",
            )

        logger.info("calculate min and max of columns")
        _temp_view_name = "temp_view_calculate_min_max_of_columns"
        clauses = []
        for column in self.columns:
            clauses.append(f"MIN({column}) AS __min__{column}")
            clauses.append(f"MAX({column}) AS __max__{column}")
        clauses = ", ".join(clauses)

        df.createOrReplaceTempView(_temp_view_name)
        collected_data = spark.sql(
            f"SELECT {clauses} FROM {_temp_view_name}"
        ).collect()
        collected_data = collected_data[0].asDict()

        results_dict = {}
        for k, v in collected_data.items():
            _column = k.replace("__min__", "").replace("__max__", "")
            _key_dict = "min_value" if "min" in k else "max_value"
            _value = v
            if isinstance(_value, datetime):
                _value = dt2str(
                    _value,
                    sep="T",
                    timespec="milliseconds",
                    add_utc_suffix=True,
                )

            updated_dict = {_key_dict: _value}
            if _column in results_dict.keys():
                old_dict = results_dict[_column]
                updated_dict.update(old_dict)
            results_dict[_column] = updated_dict

        results = []
        for k, v in results_dict.items():
            v.update({"column": k})
            results.append(v)
            logger.info(
                "column: %s, min value: %s, max value: %s",
                k,
                v.get("min_value"),
                v.get("max_value"),
            )
        result = {"results": results}
        return result


class DetectSchemaChangeWithSink(BaseSparkTransform):
    op: Literal["detect_schema_change_with_sink"]
    sink_type: str | None = None
    db_name: str | None = None
    table_name: str | None = None
    allow_table_not_exist: bool = False

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyApplyOutput:
        """Detect schema change between DF and table schema."""
        from pyspark.sql.types import ArrayType, StructField

        sql: str = "SELECT * FROM {database}.{table_name} LIMIT 0"
        logger.info("detect schema change with sink")
        if self.sink_type is not None:
            logger.info("use sink's configuration from model")
            db_name = self.db_name
            table_name = self.table_name
        else:
            logger.info("use sink's configuration from config dict")
            _sink: Sink = engine["engine"].sink
            if _sink.type in ("iceberg",):
                db_name = _sink.database
                table_name = _sink.table_name
            else:
                raise NotImplementedError("support sink only iceberg")

        if self.allow_table_not_exist is True and not is_table_exist(
            spark=spark,
            database=db_name,
            table_name=table_name,
        ):
            logger.info("skip detect schema change due to table does not exist")
            return {
                "columns_to_drop": [],
                "columns_to_add": [],
                "columns_to_alter": [],
            }

        sql = sql.format(database=db_name, table_name=table_name)
        changes = evaluate_schema_change(
            src_schema=df.schema, tgt_schema=spark.sql(sql).schema
        )
        columns_to_drop = []
        columns_to_add = []
        columns_to_alter = []
        for change in changes:
            change_type = change["type"]
            source_struct_type = change["source_struct_type"]
            parent_column = (
                ".".join(change["parent_cols"])
                if len(change["parent_cols"]) > 0
                else None
            )

            if change_type == "add_col":
                columns_to_add.append(
                    {
                        "column_name": source_struct_type.name,
                        "dtype": source_struct_type.dataType.simpleString(),
                        "add_after_column": change["add_col_after"],
                        "parent_column": parent_column,
                    }
                )
            elif change_type == "drop_col":
                columns_to_drop.append(
                    {
                        "column_name": source_struct_type.name,
                        "dtype": source_struct_type.dataType.simpleString(),
                        "parent_column": parent_column,
                    }
                )
            elif change_type == "alter_col":
                _parent_cols = change["parent_cols"]
                from_dtype = get_simple_str_dtype(change["target_struct_type"])
                to_dtype = get_simple_str_dtype(source_struct_type)

                if isinstance(source_struct_type, ArrayType) or not isinstance(
                    source_struct_type, StructField
                ):
                    column_name = _parent_cols.pop()
                    if column_name == ENUM_EXTRACT_ARRAY_TYPE:
                        column_name = _parent_cols.pop()
                else:
                    column_name = source_struct_type.name

                if len(_parent_cols) == 0:
                    _parent_cols = None

                columns_to_alter.append(
                    {
                        "column_name": column_name,
                        "from_dtype": from_dtype,
                        "to_dtype": to_dtype,
                        "parent_column": _parent_cols,
                    }
                )

        if len(columns_to_drop) > 0:
            logger.info(
                "found missing %s columns from dataframe", len(columns_to_drop)
            )
            for text in columns_to_drop:
                logger.info(text)

        if len(columns_to_add) > 0:
            logger.info(
                "found %s new columns in dataframe", len(columns_to_add)
            )
            for text in columns_to_add:
                logger.info(text)

        if len(columns_to_alter) > 0:
            logger.info(
                "found %s columns have data type's change",
                len(columns_to_alter),
            )
            for text in columns_to_alter:
                logger.info(text)
        return {
            "columns_to_drop": columns_to_drop,
            "columns_to_add": columns_to_add,
            "columns_to_alter": columns_to_alter,
        }



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, Literal

from pydantic import Field
from pydantic.functional_validators import field_validator, model_validator
from typing_extensions import Self

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, Row, SparkSession

from .....__types import DictData
from .....models import MetricSink
from .....utils import (
    clean_string,
    dt2str,
    format_bytes_humanreadable,
    regex_by_group,
    spark_env,
)
from ....__abc import BaseSink, Shape
from ...utils import (
    is_column_has_null,
    is_table_exist,
    is_timestamp_column,
    validate_col_allow_snake_case,
)
from .evolve import Evolver, EvolverModel
from .maintenance import TableMaintenance
from .utils import (
    SPARK_TEMP_VIEW_NAME,
    get_boundary_filter_condition,
    get_current_snapshot_id,
    get_hive_columns_from_df,
    get_iceberg_partition_transform_func,
    get_table_partition_columns,
)

logger = logging.getLogger("jett")


class Iceberg(BaseSink):
    """Iceberg Spark Sink model.

    :param merge_key: when mode is merge, merge key is the join key for upsert, merge key must be unique, and merge key can be many
    :param merge_sql_condition: when mode is merge, using sql condition for upsert
        it's required to use provided alias, where alias t is the target table and alias is source dataframe
        e.g. ON t.id = s.id AND t.sub_id = s.sub_id
             WHEN MATCHED AND s.op = 'delete' THEN DELETE
             WHEN MATCHED AND t.count IS NULL AND s.op = 'increment' THEN UPDATE SET t.count = 0
             WHEN MATCHED AND s.op = 'increment' THEN UPDATE SET t.count = t.count + 1
    :param partition_by: a list of columns to be partitioned, partitioning will respect the order of list
    :param table_property: a key-pair of Iceberg table property
    :param retention_days: a number of days that we will retain the data in the table
    :param retention_key: a timestamp column which will be used to filter data to be deleted
    :param keep_snapshot_days: a number of days that we will keep the snapshots,
        the snapshots that day older than param will be deleted
    :param keep_snapshot_at_least: a number of snapshots which will be retained after snapshot's expiration procedure
    :param schema_evolve_behavior: mode for table schema evolution, option is trust_source, merge, and trust_sink
    :param find_last_record_from_column: a timestamp column and the column must be first level column,
        if supply, the writer finds the maximum value of given column and put into the metric collector
        this computes after writer completes the write operation
    :param validate_column_snake_case: a boolean flag, if True, it validates that all columns in
        the DataFrame must be snake case
    :param allow_evolve_with_missing_columns: a boolean flag, if True, the workflow logic will apply table schema into
        given dataframe before writing to the target table
    :param allow_cast_safe_dtype: a boolean flag, if True, it allows to change data type of compatible columns
        (e.g. int (column in df) can be converted into double (column in table))
    """

    type: Literal["iceberg"]
    mode: Literal["append", "overwrite", "merge"] = Field(
        default="overwrite",
        description=(
            "A write mode that should be `append`, `overwrite`, or `merge` "
            "only."
        ),
    )
    database: str = Field(
        pattern=r"^[a-z0-9_]*$", description="A database name."
    )
    table_name: str = Field(
        pattern=r"^[a-z0-9_]*$", description="A table name."
    )
    table_property: dict | None = Field(default_factory=dict)
    retention_days: int | None = None
    retention_key: str | None = None
    keep_snapshot_days: int | None = 30
    keep_snapshot_at_least: int | None = 7
    find_last_record_from_column: str | None = None
    validate_column_snake_case: bool = True
    merge_key: list[str] | str | None = None
    merge_sql_condition: str | None = None

    partition_by: list[str] = Field(default_factory=list)
    schema_evolve_behavior: Literal["trust_source", "append", "trust_sink"]
    allow_evolve_with_missing_columns: bool = False
    allow_cast_safe_dtype: bool = False

    @field_validator("retention_days", mode="after")
    def __retention_period(cls, value: int | None) -> int | None:
        if value is not None and value < 1:
            raise ValueError("The `retention_days` should more than 0.")
        return value

    @model_validator(mode="after")
    def __necessary_fields(self) -> Self:
        """Validate necessary fields for this Iceberg Sink model."""
        if self.mode == "merge":
            if self.merge_key is None and self.merge_sql_condition is None:
                raise ValueError(
                    "merge mode requires merge_key or merge_sql_condition"
                )

            if (
                self.merge_key is not None
                and self.merge_sql_condition is not None
            ):
                raise ValueError(
                    "Please choose either merge_key or merge_sql_condition"
                )

        if self.retention_days is not None and self.retention_key is None:
            raise ValueError("Please input a timestamp column")

        if (
            self.schema_evolve_behavior != "append"
            and self.allow_evolve_with_missing_columns
        ):
            raise ValueError(
                "iceberg evolver `allow_evolve_with_missing_columns` option "
                "only use with schema_evolve_behavior `append`."
            )
        return self

    def validate_partition_exists(self, df: DataFrame) -> None:
        """Validate the partition column, it can be only first level column,
        not support struct

        Args:
            df (DataFrame): A Spark DataFrame.
        """
        if self.partition_by:
            first_columns_lv: list[str] = [
                c.name
                for c in df.schema
                if "struct" not in c.simpleString().split(":")[1]
            ]
            partition_cols: list[str] = []
            for p in self.partition_by:
                # NOTE: Start Prepare string value if it passes with bucket like
                #   `["bucket(2, _id)"]`.
                #   - Step: "bucket(2, _id)" -> "2, _id)"
                #   - Step: "2, _id)"        -> "2, _id"
                #   - Step: "2, _id"         -> " _id"
                #   - Step: " _id"           -> "_id"
                #
                _p: str = p.split("(")[-1]
                _p: str = _p.replace(")", "")
                _p: str = _p.split(",")[-1]
                _p: str = _p.strip()
                partition_cols.append(_p)

            for p in partition_cols:
                if p not in first_columns_lv:
                    raise ValueError(
                        f"Partition column {p} not found, partition column "
                        f"must be 1st level column only"
                    )

    def _extract_merge_key(self, df) -> list[str]:
        """Get merge key from component model."""
        merge_keys = None
        if self.merge_key is not None:
            merge_keys = (
                [self.merge_key]
                if isinstance(self.merge_key, str)
                else self.merge_key
            )
        elif self.merge_sql_condition is not None:
            on_clause = re.split(
                r"(?i)when",
                re.split(r"(?i)on", self.merge_sql_condition, maxsplit=1)[1],
                maxsplit=1,
            )[0].strip()
            merge_keys = re.findall(r"s\.(\w+)", on_clause)

        missing_keys = [key for key in merge_keys if key not in df.columns]
        if not merge_keys or missing_keys:
            raise ValueError(
                f"please check source dataframe, there are missing merge keys: {missing_keys}"
            )
        return merge_keys

    def _validate_duplicate_merge_key(self, df: DataFrame) -> None:
        """
        Validate that there are no duplicate values in the merge_key columns
        """
        if self.mode != "merge":
            return

        merge_keys = self._extract_merge_key(df=df)

        duplicate_df = df.groupBy(*merge_keys).count().filter("count > 1")
        if not duplicate_df.isEmpty():
            raise ValueError(
                f"Duplicate values found for merge keys: {merge_keys}"
            )

    def _get_table_property(self) -> str:
        """
        get and update default table property
        """
        table_property = self._default_table_property
        if len(self.table_property) > 0:
            for k, v in self.table_property.items():
                table_property[k] = v

        table_property = [f"'{k}'='{v}'" for k, v in table_property.items()]
        table_property = ",\n".join(table_property)
        return table_property

    def get_partition_ddl(self) -> str:
        """Get partition by clause."""
        if self.partition_by:
            return f"PARTITIONED BY ({', '.join(self.partition_by)})"
        return ""

    def get_table_ddl(self, df: DataFrame) -> str:
        """
        get iceberg table ddl
        """
        columns = get_hive_columns_from_df(df=df)
        table_property = self._get_table_property()
        partition_by = self.get_partition_ddl()
        ddl = self._create_table_stmt.format(
            hive_db=self.database,
            table_name=self.table_name,
            columns=columns,
            partition_by=partition_by,
            properties=table_property,
        )
        ddl = clean_string(text=ddl)
        return ddl

    def _create_database(self, spark: SparkSession) -> None:
        """
        create database if not exist, for local environment only
        aws env and iu env must already create the Hive DB
        """
        if spark_env("ENV") == "local":
            if not spark.catalog.databaseExists(dbName=self.database):
                # NOTE: no need to define location since it's already defined
                #   in spark-warehouse.
                ddl: str = f"CREATE DATABASE {self.database}"
                logger.info("database does not exists, create a database")
                logger.info("execute sparksql: %s", ddl)
                spark.sql(ddl)
        else:
            if not spark.catalog.databaseExists(dbName=self.database):
                raise RuntimeError(
                    "please create database in your env (aws / iu) before write a table"
                )

    def _create_table(
        self, df: DataFrame, spark: SparkSession, metric: MetricSink
    ) -> None:
        """
        create iceberg table if not exist
        """
        if not is_table_exist(
            spark=spark,
            database=self.database,
            table_name=self.table_name,
        ):
            ddl = self.get_table_ddl(df=df)
            logger.info("table does not exist, create a new iceberg table")
            logger.info(f"execute sparksql: {ddl}")
            spark.sql(ddl)
            self._is_table_created = True

            # NOTE: set metric sink of table operation when table is created.
            operations = []
            for schema in df.schema.fields:
                _s = schema.simpleString().split(":", maxsplit=1)
                operations.append(
                    {
                        "name": _s[0],
                        "data_type": _s[1],
                    }
                )
            metric.add(
                key="table_operation_create_table_columns",
                value=operations,
            )
        else:
            logger.info("table exists")
            self._is_table_created = False

    def _get_append_statement(self) -> str:
        """
        get SparkSQL insert into statement
        """
        sql = f"INSERT INTO {self.database}.{self.table_name} SELECT * FROM {SPARK_TEMP_VIEW_NAME}"
        sql = clean_string(text=sql)
        return sql

    def _get_overwrite_statement(self, spark: SparkSession) -> str:
        """
        get SparkSQL insert overwrite statement
        """
        spark.sql("SET spark.sql.sources.partitionOverwriteMode=dynamic")
        sql = f"INSERT OVERWRITE {self.database}.{self.table_name} SELECT * FROM {SPARK_TEMP_VIEW_NAME}"
        sql = clean_string(text=sql)
        return sql

    def _get_merge_table_filter_condition(
        self, df: DataFrame, spark: SparkSession
    ) -> str:
        """
        get filter condition using in MERGE operation
        help reducing data scanning
        """
        allow_partition_transform_func = [
            "year",
            "month",
            "day",
            "hour",
        ]

        table_filter = None
        partition_spec = get_table_partition_columns(
            spark=spark,
            db_name=self.database,
            table_name=self.table_name,
        )
        if len(partition_spec) > 0:
            # assume that 1st partition column must be a timestamp column, if not it cannot apply filter
            first_partition_spec = partition_spec[0]
            transform_func = get_iceberg_partition_transform_func(
                partition_spec=first_partition_spec
            )
            if transform_func in allow_partition_transform_func:
                partition_col = regex_by_group(
                    text=first_partition_spec,
                    regex=r"(?<=\()[^)]+(?=\))",
                    n=0,
                )  # get everything inside ()

                logger.info(
                    "found 1st partition column that is a timestamp column: %s",
                    first_partition_spec,
                )
                logger.info("get timestamp filter for merge operation")
                sql = f"""
                    SELECT MIN({partition_col}) AS min_val, MAX({partition_col}) AS max_val
                    FROM {SPARK_TEMP_VIEW_NAME} WHERE {partition_col} IS NOT NULL
                """
                data_row = spark.sql(sql).collect()[0]
                logger.info("min value: %s", data_row.min_val)
                logger.info("max value: %s", data_row.max_val)

                if data_row.min_val is None or data_row.max_val is None:
                    logger.info(
                        "There are all None in partition column `%s`. So we will filter target table in null partition only ",
                        partition_col,
                    )
                    table_filter = f"t.{partition_col} IS NULL"
                    return table_filter

                min_val, max_val = get_boundary_filter_condition(
                    transform_func=transform_func,
                    start=data_row.min_val,
                    end=data_row.max_val,
                )

                table_filter = f'(t.{partition_col} >= "{min_val}" AND t.{partition_col} < "{max_val}")'

                # check that df contains null or not
                is_col_has_null = is_column_has_null(
                    df=df, column=partition_col
                )
                if is_col_has_null:
                    table_filter = (
                        f"({table_filter} OR t.{partition_col} IS NULL)"
                    )

        return table_filter

    def _get_merge_condition(self, df: DataFrame, spark: SparkSession) -> str:
        """
        get merge condition
        """
        merge_condition = self.merge_sql_condition
        if self.merge_key is not None:
            _merge_key = self.merge_key
            if isinstance(self.merge_key, str):
                _merge_key = [self.merge_key]
            merge_clauses = [f"t.{k} = s.{k}" for k in _merge_key]
            if not self._is_table_created:
                table_filter = self._get_merge_table_filter_condition(
                    df, spark=spark
                )
                if table_filter is not None:
                    merge_clauses = merge_clauses + [table_filter]
            on_condition = " AND ".join(merge_clauses)
            merge_condition = f"""
            ON {on_condition}
            WHEN MATCHED THEN UPDATE SET *
            WHEN NOT MATCHED THEN INSERT *
            """
        merge_condition = clean_string(merge_condition)
        return merge_condition

    def _get_merge_statement(self, df: DataFrame, spark: SparkSession) -> str:
        """
        get SparkSQL merge into statement
        """
        merge_condition = self._get_merge_condition(df=df, spark=spark)
        sql = f"""
        MERGE INTO {self.database}.{self.table_name} AS t
        USING (SELECT * FROM {SPARK_TEMP_VIEW_NAME}) AS s
        {merge_condition}
        """
        sql = clean_string(sql)
        return sql

    def _get_write_sql_statement(
        self, df: DataFrame, spark: SparkSession
    ) -> str:
        sql = None
        logger.info("iceberg write mode: %s", self.mode)
        match self.mode:
            case "append":
                sql = self._get_append_statement()
            case "overwrite":
                sql = self._get_overwrite_statement(spark=spark)
            case "merge":
                sql = self._get_merge_statement(df, spark=spark)
        return sql

    def _get_write_statistics(self, spark: SparkSession) -> dict:
        """Get write statistics of last table write operation."""
        snapshot_id: str = get_current_snapshot_id(
            spark=spark, database=self.database, table_name=self.table_name
        )
        raw_stats: Row = spark.sql(
            f"SELECT * FROM {self.database}.{self.table_name}.snapshots "
            f"WHERE snapshot_id = '{snapshot_id}'"
        ).collect()[0]
        added_file_size = raw_stats.summary.get("added-files-size", 0)
        stats = {
            "written_rows": raw_stats.summary.get("added-records", 0),
            "written_bytes": added_file_size,
            "written_bytes_formatted": (
                format_bytes_humanreadable(int(added_file_size))
                if added_file_size is not None
                else None
            ),
            "affected_partitions": raw_stats.summary.get(
                "changed-partition-count", 0
            ),
        }
        return stats

    def _find_last_record_from_column(
        self,
        spark: SparkSession,
        metric: MetricSink,
    ) -> None:
        """Compute last record of given input and set the metric."""
        if self.find_last_record_from_column is not None:
            table_schema = spark.sql(
                f"SELECT * FROM {self.database}.{self.table_name} LIMIT 0"
            ).schema
            is_timestamp_col = is_timestamp_column(
                column_name=self.find_last_record_from_column,
                schema=table_schema,
            )
            if not is_timestamp_col:
                raise ValueError(
                    f"{self.find_last_record_from_column} is not a timestamp "
                    f"column",
                )

            logger.info(
                "find last record from column: %s",
                self.find_last_record_from_column,
            )
            find_last = f"""
                SELECT MAX(
                    readable_metrics.{self.find_last_record_from_column}.upper_bound
                )
                FROM {self.database}.{self.table_name}.files
            """
            find_last = clean_string(find_last)
            logger.info(find_last)
            last_record = spark.sql(find_last).first()[0]
            last_record = dt2str(last_record, sep=" ", timespec="microseconds")

            logger.info("last record: %s", last_record)
            metric.add("latest_record_update", last_record)
            metric.add(
                "latest_record_update_from_column",
                self.find_last_record_from_column,
            )

    def _apply_table_schema_to_dataframe(
        self, df: DataFrame, spark: SparkSession
    ) -> DataFrame:
        if self.allow_evolve_with_missing_columns:
            current_table_schema = spark.sql(
                f"SELECT * FROM {self.database}.{self.table_name} LIMIT 0"
            ).schema
            if df.schema != current_table_schema:
                logger.info("apply table schema to dataframe")
                df = df.to(current_table_schema)

        return df

    def save(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricSink,
        **kwargs,
    ) -> Any:
        """Save the result data to the Iceberg table.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData):
            metric (MetricSink):
        """
        logger.info("Sink - Start sink to Iceberg.")
        spark: SparkSession = engine["spark"]

        if df.isEmpty():
            logger.warning("Sink - Found empty dataframe, skip write")
            metric.add("affected_partitions", 0)
            return df, Shape(rows=0, columns=0)

        logger.info("Sink - Validate table spec and dataframe")
        if self.validate_column_snake_case:
            validate_col_allow_snake_case(schema=df.schema)
        self.validate_partition_exists(df=df)

        logger.info(f"> database name: {self.database}")
        logger.info(f"> table name: {self.table_name}")

        logger.info("Start Iceberg Table Evolve ...")
        evolve_model = EvolverModel.model_validate(self, from_attributes=True)
        evolver = Evolver(model=evolve_model, df=df, spark=spark)

        logger.info("start table maintenance ...")
        maintain = TableMaintenance.model_validate(self, from_attributes=True)
        maintain.validate_retention_key(df)

        # start Iceberg write operation
        logger.info("write df to iceberg table")
        self._create_database(spark=spark)
        self._create_table(df, spark=spark, metric=metric)

        evolver.evolve_schema()
        evolver.evolve_partition()
        df = evolver.df

        metric.add("alter_history", evolver.list_sql_alter)
        alter_list, add_list, drop_list = evolver.get_simple_detail_changes()
        metric.add("table_operation_alter_columns", alter_list)
        metric.add("table_operation_add_columns", add_list)
        metric.add("table_operation_drop_columns", drop_list)

        df = self._apply_table_schema_to_dataframe(df, spark=spark)

        self._validate_duplicate_merge_key(df=df)
        df.createOrReplaceTempView(SPARK_TEMP_VIEW_NAME)
        sql = self._get_write_sql_statement(df=df, spark=spark)
        logger.info("start writing data into table")
        logger.info(f"Sink - Execute sparksql: {sql}")
        spark.sql(sql)
        logger.info("wrote successfully")

        stats = self._get_write_statistics(spark=spark)
        logger.info(f"> Written rows: {stats['written_rows']}")
        logger.info(
            f"> Written bytes: {stats['written_bytes']} "
            f"({stats['written_bytes_formatted']})"
        )
        logger.info(
            f"> Written Affected partitions: {stats['affected_partitions']}"
        )
        metric.add("affected_partitions", int(stats.get("affected_partitions")))

        maintain.retain_data(spark=spark)
        maintain.expire_snapshot(spark=spark)

        # NOTE: Write operation already completed
        self._find_last_record_from_column(spark=spark, metric=metric)
        return (
            df,
            Shape(rows=stats.get("written_rows"), columns=len(df.columns)),
        )

    def outlet(self) -> tuple[str, str]:
        return "iceberg", self.dest()

    def dest(self) -> str:
        return f"{self.table_name}"



from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal

from ....__types import DictData
from ....models import Context, MetricOperator
from ...__abc import BaseTransform

if TYPE_CHECKING:
    from duckdb import DuckDBPyRelation as Relation
    from duckdb.experimental.spark.sql import DataFrame, SparkSession
    from duckdb.experimental.spark.sql.column import Column

    PairCol = tuple[Column, str]

logger = logging.getLogger("jett")


def is_pair_col(value: PairCol | Any) -> bool:
    """Change value is a pair of Column and string."""
    return (
        isinstance(value, tuple)
        and len(value) == 2
        and isinstance(value[0], Column)
        and isinstance(value[1], str)
    )


class BaseDuckDBTransform(BaseTransform, ABC):
    """Base DuckDB Transform abstract model."""

    priority: Literal["pre", "group", "post"] = "pre"

    @abstractmethod
    def apply(
        self,
        df: Relation | DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> Relation | DataFrame:
        """Apply priority transform."""

    def apply_group(
        self,
        df: Relation | DataFrame,
        engine: DictData,
        metric: MetricOperator,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> str | PairCol | list[PairCol]:
        """Apply group transform."""
        raise NotImplementedError(
            f"Transform: {self.__class__.__name__} on DuckDB engine does not "
            f"implement the group operation, please change this to priority "
            f"operator."
        )

    def handle_apply_group(
        self,
        df: Relation | DataFrame,
        context: Context,
        engine: DictData,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> Any:
        ts: float = time.monotonic()
        metric = MetricOperator(type="order", trans_op=self.op)
        logger.info(f"🔨 Handle Apply Group Operator: {self.op!r}")
        output: str | PairCol | list[PairCol] = self.apply_group(
            df, engine, spark=spark, metric=metric, **kwargs
        )
        if is_pair_col(output):
            rs: dict[str, Column] = {output[1]: output[2]}
        else:
            raise NotImplementedError()
        metric.transform_latency_ms = time.monotonic() - ts
        context["metric_group_transform"].transforms.append(metric)
        return rs

    @staticmethod
    def sync_schema(pre, post, metric, **kwargs) -> None: ...



from typing import Annotated, Union

from pydantic import Field

from .files.gcs import GCSCSVFile, GCSJsonFile
from .files.local import LocalCSVFile, LocalJsonFile
from .files.s3 import S3CSVFile, S3JsonFile
from .hive import Hive
from .jdbc import Jdbc, MySQL, Postgres
from .mongodb import MongoDB

LocalFile = Annotated[
    Union[
        LocalCSVFile,
        LocalJsonFile,
    ],
    Field(discriminator="file_format"),
]

S3File = Annotated[
    Union[
        S3CSVFile,
        S3JsonFile,
    ],
    Field(discriminator="file_format"),
]

GCSFile = Annotated[
    Union[
        GCSCSVFile,
        GCSJsonFile,
    ],
    Field(discriminator="file_format"),
]


Source = Annotated[
    Union[
        LocalFile,
        S3File,
        GCSFile,
        MongoDB,
        Postgres,
        MySQL,
        Jdbc,
        Hive,
    ],
    Field(
        discriminator="type",
        description="The source registry.",
    ),
]



from typing import Any, Literal

from ....__types import DictData
from ....models import MetricSource, Shape
from ...__abc import BaseSource


class RestApi(BaseSource):
    type: Literal["restapi"]
    host: str

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[Any, Shape]: ...

    def inlet(self) -> tuple[str, str]: ...



from __future__ import annotations

import json
import logging
from typing import Literal

from pydantic import Field

from ..__types import DictData
from ..models import MetricData
from .__abc import BaseMetric

logger = logging.getLogger("jett")


class Console(BaseMetric):
    """Console Metric model that use to emit metric to logging with a specific
    logging name.
    """

    type: Literal["console"] = Field(description="A type of console metric.")
    name: str = Field(default="jett", description="A logging name.")

    def emit(self, data: MetricData, convert: DictData, **kwargs) -> None:
        """Emit Console method.

        Args:
            data (MetricData): A MetricData model that passing after end the
                engine execution.
            convert (DictData): A converted data from the metric data.
        """
        logger.info("📩 Metric - console")
        console = logging.getLogger(self.name)
        console.info(
            f"Console Metric:\n{json.dumps(convert, default=str, indent=1)}"
        )



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Final

from dateutil.relativedelta import relativedelta

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, Row, SparkSession
    from pyspark.sql.connect.session import SparkSession as SparkRemoteSession
    from pyspark.sql.types import (
        ArrayType,
        DataType,
        StructField,
        StructType,
    )

    AnyType = StructType | ArrayType | StructField | DataType
    Changed = dict[str, str | AnyType | list[str]]

from .....utils import regex_by_group
from ...utils import ENUM_EXTRACT_ARRAY_TYPE

logger = logging.getLogger("jett")
ICEBERG_HIDDEN_PARTITION_FUNC_MAPPING: Final[dict[str, str]] = {
    "years": "year",
    "year": "year",
    "months": "month",
    "month": "month",
    "days": "day",
    "day": "day",
    "date": "day",
    "hours": "hour",
    "hour": "hour",
    "date_hour": "hour",
}
CREATE_TABLE_STMT: Final[
    str
] = """
CREATE EXTERNAL TABLE `{database}`.`{table_name}`(
    {columns}
)
USING ICEBERG
    {partition_by}
TBLPROPERTIES (
    {properties}
)
"""
DEFAULT_TABLE_PROPERTY: Final[dict[str, str]] = {
    "format-version": "2",
    "write.format.default": "parquet",
    "write.distribution-mode": "hash",
    "write.parquet.compression-codec": "zstd",
    "write.avro.compression-codec": "zstd",
    "write.orc.compression-codec": "zstd",
    "write.target-file-size-bytes": "524288000",
    "engine.hive.enabled": "TRUE",
}
SPARK_TEMP_VIEW_NAME: Final[str] = "tool_temp_table"

ListStr = list[str]


def get_iceberg_partition_transform_func(partition_spec: str) -> str:
    """
    extract iceberg's partition transform function
    e.g. days(created_at) -> return as days
    """
    _split: list[str] = partition_spec.split("(", maxsplit=1)
    transform_func: str = ""
    if len(_split) > 1:
        transform_func: str = _split[0]
        if transform_func in ICEBERG_HIDDEN_PARTITION_FUNC_MAPPING:
            transform_func = ICEBERG_HIDDEN_PARTITION_FUNC_MAPPING[
                transform_func
            ]
    return transform_func


def get_table_partition_columns(
    spark: SparkSession | SparkRemoteSession, db_name: str, table_name: str
) -> list[str]:
    """Get partition columns from Iceberg table."""
    partition_spec = []
    rows: list[Row] = spark.sql(
        f"DESC EXTENDED {db_name}.{table_name}"
    ).collect()
    for row in rows:
        if _match := regex_by_group(text=row.col_name, regex=r"Part\s\d+", n=0):
            partition_spec.append(row.data_type)
    return partition_spec


def get_current_snapshot_id(
    spark: SparkSession | SparkRemoteSession, database: str, table_name: str
) -> str:
    """Get the current snapshot id from DDL."""
    rows: list[Row] = spark.sql(
        f"DESC EXTENDED {database}.{table_name}"
    ).collect()
    table_properties: list[str] = [
        row.data_type for row in rows if row.col_name == "Table Properties"
    ]
    # NOTE: Empty table also has table property where value is set to
    #   `current-snapshot-id=none`.
    table_properties: str = table_properties[0]
    snapshot_id: str = regex_by_group(
        text=table_properties, regex=r"current-snapshot-id=(\d+)", n=1
    )
    return snapshot_id


def extract_hidden_partition(partition_spec: list[str]) -> dict[str, str]:
    """Extract Iceberg hidden partitioning spec and function.

    Args:
        partition_spec (list[str]): list of partition specification
            (e.g. ['day(created_at)', 'category'])

    Returns:
        dict: dict of partition specification, the example structure is shown as below
            {
                'created_at': 'years(created_at)'
                'category': 'category',
                'id': 'bucket(4, id)',
                'category': 'truncate(10, category)',
            }
    """
    rs = {}
    for p in partition_spec:
        is_hidden_partition: bool = False
        extracted_col: str = regex_by_group(
            text=p, regex=r"(?<=\()[^)]+(?=\))", n=0
        )  # get everything inside ()

        if extracted_col != "":
            is_hidden_partition = True
            extracted_col = extracted_col.split(",")[-1].strip()

        column: str = extracted_col if is_hidden_partition else p
        rs[column] = p
    return rs


def clean_hidden_partition(partition_spec: str) -> str:
    """Unify iceberg hidden partition transformation supports legacy case like
    days == day.
    """
    _splits: list[str] = partition_spec.replace(" ", "").split("(")
    _key = _splits[0]
    if _key in ICEBERG_HIDDEN_PARTITION_FUNC_MAPPING.keys():
        _val = _splits[1]
        hidden_func = ICEBERG_HIDDEN_PARTITION_FUNC_MAPPING[_key]
        partition_spec = f"{hidden_func}({_val}"
    return partition_spec


def gen_alter_cols(changes: list[Changed]) -> tuple[ListStr, ListStr, ListStr]:
    """Classify and generate the list of columns for add, alter, and drop.

    Returns:
        ListStr: List of new columns, which will be added into table
        ListStr: List of columns, which have type changed
        ListStr: List of columns, which will be dropped from table
    """
    to_add_col_cmds: list[str] = []
    to_alter_col_cmds: list[str] = []
    to_drop_col_cmds: list[str] = []

    for change in changes:
        change_type: str = change.get("type")
        parent_cols: list[str] = change.get("parent_cols")
        add_col_after: str | None = change.get("add_col_after", None)
        source_struct_splits: list[str] = (
            change.get("source_struct_type").simpleString().split(":")
        )
        source_name: str = source_struct_splits[0]
        source_dtype: str = ":".join(source_struct_splits[1:])
        parent_cols_str: str = ".".join(parent_cols)

        if len(parent_cols) != 0:
            source_name: str = f"{parent_cols_str}.{source_name}"

        source_name = source_name.replace(ENUM_EXTRACT_ARRAY_TYPE, "element")

        if change_type == "add_col":
            add_col_cmd = f"{source_name} {source_dtype}"
            if add_col_after:
                add_col_cmd = f"{add_col_cmd} AFTER {add_col_after}"
            else:
                add_col_cmd = f"{add_col_cmd} FIRST"
            to_add_col_cmds.append(add_col_cmd)

        elif change_type == "alter_col":
            to_alter_col_cmds.append(f"{source_name} TYPE {source_dtype}")

        elif change_type == "drop_col":
            to_drop_col_cmds.append(source_name)

    return to_add_col_cmds, to_alter_col_cmds, to_drop_col_cmds


def gen_alter_partitions(
    database: str,
    table_name: str,
    source_partition: dict,
    table_partition: dict,
) -> tuple[list[str], list[str], list[str]]:
    """
    compare what partition specification is needed to drop, replace, or add
    Args:
        database:
        table_name:
        source_partition: a dict, that contains the information of partition specification of configuration
        table_partition: a dict, that contains the information of partition specification of table

    Returns:
        list[str]: List of commands for adding new partitions
        list[str]: List of commands for replacing existing partitions
        list[str]: List of commands for dropping existing partitions
    """

    to_add = []
    to_replace = []
    to_drop = []

    source_keys = source_partition.keys()
    table_keys = table_partition.keys()

    for k, v in source_partition.items():
        if k not in table_keys:  # partition not exist, add new partition
            to_add.append(f"ADD PARTITION FIELD {v}")
        else:
            table_partition_spec = clean_hidden_partition(table_partition[k])
            source_partition_spec = clean_hidden_partition(v)

            if (
                source_partition_spec != table_partition_spec
            ):  # partition spec is diff, replace partition
                to_replace.append(
                    f"REPLACE PARTITION FIELD {table_partition_spec} WITH {v}"
                )

    for k, v in table_partition.items():
        if (
            k not in source_keys
        ):  # partition not exist in configuration, drop existing partition
            to_drop.append(f"DROP PARTITION FIELD {v}")

    to_add = [f"ALTER TABLE {database}.{table_name} {t}" for t in to_add]
    to_replace = [
        f"ALTER TABLE {database}.{table_name} {t}" for t in to_replace
    ]
    to_drop = [f"ALTER TABLE {database}.{table_name} {t}" for t in to_drop]

    return to_add, to_replace, to_drop


def get_boundary_filter_condition(
    transform_func: str, start: datetime, end: datetime
) -> tuple[str, str]:
    """Calculate a new range of start and end based on Iceberg partition
    transform function
    Examples:
        - transform_func is day
        - start is 2024-01-10 12:00:00
        - end is 2024-01-11 17:30:00

        the new range will be 2024-01-10 00:00:00 and 2024-01-12 00:00:00
    """
    mapping = {
        "year": {
            "format": "%Y-01-01 00:00:00",
            "delta": relativedelta(years=1),
        },
        "month": {
            "format": "%Y-%m-01 00:00:00",
            "delta": relativedelta(months=1),
        },
        "day": {
            "format": "%Y-%m-%d 00:00:00",
            "delta": relativedelta(days=1),
        },
        "hour": {
            "format": "%Y-%m-%d %H:00:00",
            "delta": relativedelta(hours=1),
        },
    }
    val = mapping[transform_func]
    min_val_str = (start - val["delta"]).strftime(val["format"])
    max_val_str = (end + val["delta"]).strftime(val["format"])
    return min_val_str, max_val_str


def get_hive_columns_from_df(df: DataFrame) -> str:
    """Get name of columns with data type, compatible with hive column."""
    columns = []
    for _, schema in enumerate(df.schema.fields):
        raw_schema = schema.simpleString()
        _s = raw_schema.split(":", maxsplit=1)
        columns.append(f"`{_s[0]}` {_s[1]}")
    return ",\n".join(columns)


def is_column_safe_to_alter(
    current_dtype: StructField,
    to_be_dtype: StructField,
) -> bool:
    """Check compatibility of changing data type for Iceberg.

    References:
        - https://iceberg.apache.org/docs/latest/spark-ddl/#alter-table-alter-column
    """
    from pyspark.sql.types import (
        DecimalType,
        DoubleType,
        FloatType,
        IntegerType,
        LongType,
    )

    if current_dtype.__class__ != to_be_dtype.__class__:
        return False

    is_safe_update = False
    _current = current_dtype.dataType
    _change = to_be_dtype.dataType
    if isinstance(_current, IntegerType) and isinstance(_change, LongType):
        is_safe_update = True
    elif isinstance(_current, FloatType) and isinstance(_change, DoubleType):
        is_safe_update = True
    elif isinstance(_current, DecimalType) and isinstance(_change, DecimalType):
        current_value = regex_by_group(
            text=_current.jsonValue(), regex=r"\d+,\d+", n=0
        )
        target_value = regex_by_group(
            text=_change.jsonValue(), regex=r"\d+,\d+", n=0
        )

        current_precision, current_scale = current_value.split(",")
        target_precision, target_scale = target_value.split(",")
        if (
            int(target_precision) > int(current_precision)
            and current_scale == target_scale
        ):
            is_safe_update = True
    return is_safe_update



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, Final, Literal

from pydantic import Field, SecretStr
from pydantic.functional_validators import field_validator, model_validator

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession

from ....__types import DictData
from ....models import BasicFilter, Shape
from ...__abc import BaseSource
from ..temp_storage import TempStorage

logger = logging.getLogger("jett")

ALLOWED_SSL_MODE: Final[tuple[str, ...]] = ("DISABLED", "PREFERRED", "REQUIRED")


class Jdbc(BaseSource):
    """JDBC Spark Source model."""

    type: Literal["jdbc"]
    protocol: str | None = None
    driver: str | None = None
    host: str = Field(description="A host URL for JDBC URI.")
    port: int | None = Field(
        default=None,
        description="A port number for JDBC URI.",
        ge=0,
    )
    username: str = Field(description="A username for JDBC URI.")
    password: SecretStr
    database: str
    table_name: str | None = None
    query: str | None = None
    properties: dict[str, str | int | bool] | None = None
    filter: list[BasicFilter] = Field(default_factory=list)
    sample_records: int | None = None
    fetch_size: int = Field(default=0, gt=0)
    query_timeout: int | None = 0
    partition_column: str | None = None
    lower_bound: int | str | None = None
    upper_bound: int | str | None = None
    num_partitions: int = Field(default=1)
    custom_schema: str | None = Field(None, alias="schema")
    allow_save_data_to_temp: bool = True
    lazy: bool = Field(default=True)

    @field_validator("database", "table_name", "partition_column")
    def __validate_space(cls, value: Any) -> Any:
        """Check field should not include space character."""
        if value and isinstance(value, str) and " " in value:
            raise ValueError("Cannot contain whitespace")
        return value

    @field_validator(*("lower_bound", "upper_bound"))
    def __check_boundary(cls, value: Any) -> Any:
        if value is None:
            return value

        if isinstance(value, str):
            try:
                datetime.fromisoformat(value)
                return value
            except ValueError:
                pass
        elif isinstance(value, int):
            return value

        raise ValueError(
            "field must be an integer, or an ISO format date/timestamp"
        )

    @model_validator(mode="after")
    def __post_validation(self):
        if (self.upper_bound is not None and self.lower_bound is None) or (
            self.upper_bound is None and self.lower_bound is not None
        ):
            raise ValueError("lower_bound or upper_bound is missing")

        if self.lower_bound is not None:
            _upper_bound = (
                self.upper_bound
                if isinstance(self.upper_bound, int)
                else datetime.fromisoformat(self.upper_bound)
            )
            _lower_bound = (
                self.lower_bound
                if isinstance(self.lower_bound, int)
                else datetime.fromisoformat(self.lower_bound)
            )
            if _upper_bound < _lower_bound:
                raise ValueError("upper bound must be greater than lower bound")

        if self.query is not None:
            if self.table_name is not None:
                raise ValueError(
                    "query and table_name cannot use at the same time"
                )
            elif self.partition_column is not None:
                raise ValueError(
                    "when query is set, partition_column is not allowed"
                )
        else:
            if self.table_name is None:
                raise ValueError("please supply either query or table_name")
        return self

    def get_tbl_or_query(self) -> dict[str, str]:
        target = {}
        if self.table_name is not None:
            target = {"dbtable": self.fullname()}
            logger.info(f"Source - table name: {self.fullname()}")
        elif self.query is not None:
            target = {"query": self.query}
            logger.info(f"Source - query: {self.query}")
        return target

    def get_uri(self) -> str:
        """Get JDBC connection uri."""
        uri = f"jdbc:{self.protocol}://{self.host}:{self.port}/{self.database}"
        if self.properties:
            parsed_properties = "&".join(
                [f"{k}={v}" for k, v in self.properties.items()]
            )
            uri = f"{uri}?{parsed_properties}"
        return uri

    def get_jdbc_options(self) -> dict[str, str]:
        """Get JDBC options."""
        uri: str = self.get_uri()
        options = {
            "driver": self.driver,
            "url": uri,
            "user": self.username,
            "password": self.password.get_secret_value(),
            "fetchSize": self.fetch_size,
            "queryTimeout": self.query_timeout,
        }
        logger.info(f"Source - uri: {uri}")
        logger.info(f"Source - username: {self.username}")
        logger.info(f"Source - JDBC driver name: {self.driver}")
        logger.info(f"fetch size: {self.fetch_size}")
        logger.info(f"query timeout: {self.query_timeout}")
        if self.custom_schema:
            options.update({"customSchema": self.custom_schema})
            logger.info(f"schema: {self.custom_schema}")
        return options

    def get_boundary(
        self, spark: SparkSession, options: dict[str, str]
    ) -> dict[str, str | int]:
        """Get the boundary if partition exists."""
        boundary_options = {"numPartitions": self.num_partitions}
        if self.partition_column is not None and self.num_partitions > 1:
            lower_bound = self.lower_bound
            upper_bound = self.upper_bound
            if (
                lower_bound is None
                and upper_bound is None
                and self.query is None
            ):
                logger.info(
                    "boundary not found. lower bound and upper bound will be "
                    "automatically updated."
                )
                tmp_options = options.copy()
                tmp_options["query"] = (
                    f"SELECT "
                    f"  MIN({self.partition_column}) "
                    f", MAX({self.partition_column}) "
                    f"FROM {self.fullname()}"
                )
                logger.info(f"query to find boundary: {tmp_options['query']}")
                lower_bound, upper_bound = (
                    spark.read.format("jdbc")
                    .options(**tmp_options)
                    .load()
                    .first()
                )

            boundary_options.update(
                {
                    "partitionColumn": self.partition_column,
                    "lowerBound": lower_bound,
                    "upperBound": upper_bound,
                }
            )

            logger.info(f"lower bound: {lower_bound}")
            logger.info(f"upper bound: {upper_bound}")
            logger.info(f"partition column: {self.partition_column}")
            logger.info(f"num partitions: {self.num_partitions}")

        return boundary_options

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        """Load JDBC to Spark DataFrame object.

        Args:
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
        """
        spark: SparkSession = engine["spark"]
        options = self.get_jdbc_options()
        options: dict[str, str | int] = (
            options
            | self.get_boundary(spark, options=options)
            | self.get_tbl_or_query()
        )
        logger.info("load data from source via jdbc")
        df: DataFrame = spark.read.format("jdbc").options(**options).load()

        # NOTE: Apply filter.
        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        if not self.allow_save_data_to_temp:
            return df, Shape()

        temp_storage = TempStorage(spark, prefix=f"source__{self.type}")
        df = temp_storage.apply(df)
        rows: int = df.count()
        logger.info(f"Source - rows: {rows}")
        shape: tuple[int, int] = (rows, len(df.columns))
        engine["temp_storage"]["source"] = temp_storage
        return df, Shape.from_tuple(shape)

    def fullname(self):
        return self.table_name

    def inlet(self) -> tuple[str, str]:
        return self.type, self.fullname()


class Postgres(Jdbc):
    """Postgres Spark Source model."""

    type: Literal["postgresql"]
    protocol: str = "postgresql"
    driver: str = "org.postgresql.Driver"
    port: int = 5432
    schema_name: str = "public"

    def fullname(self) -> str:
        return f"{self.schema_name}.{self.table_name}"


class MySQL(Jdbc):
    """
    A Configuration Model for MySQL (Spark Engine)
    Implemented on top of BaseJDBCModel

    :param protocol: a protocol of MySQL
    :param driver: a driver name of MySQL
    :param port: port number
    """

    type: Literal["mysql"]
    protocol: Literal[
        "mysql",
        "mysql:loadbalance",
        "mysql:replication",
        "mysqlx",
        "mysql+srv",
        "mysql+srv:loadbalance",
        "mysql+srv:replication",
        "mysqlx+srv",
    ] = "mysql"
    driver: str = "com.mysql.cj.jdbc.Driver"
    port: int = 3306

    @field_validator("properties")
    def __validate_properties(cls, value: Any) -> Any:
        """Validate JDBC properties for MySQL data source."""
        if not isinstance(value, dict):
            return value

        for k, v in value.items():
            if isinstance(v, str) and v in ("True", "False"):
                value[k] = v.lower()
            if k == "sslMode" and v not in ALLOWED_SSL_MODE:
                raise ValueError(
                    f"sslMode {v} not allowed, support only {ALLOWED_SSL_MODE}"
                )
        return value

    def fullname(self) -> str:
        """Return the fullname of this MySQL data source."""
        return f"{self.database}.{self.table_name}"



import logging
from typing import Literal

try:
    from daft import DataFrame

    DAFT_EXISTS: bool = True
except ImportError:
    DAFT_EXISTS: bool = False

from ....__types import DictData
from ....models import Shape
from ...__abc import BaseSink

logger = logging.getLogger("jett")


class Console(BaseSink):
    """Console DuckDB Sink model."""

    type: Literal["console"]
    limit: int = 10
    max_width: int | None = None

    def save(
        self,
        df: "DataFrame",
        *,
        engine: DictData,
        **kwargs,
    ) -> tuple["DataFrame", Shape]:
        """Save the result data to the Console."""
        # NOTE: Convert variable from df to table.
        logger.info("🎯 Sink - Start sync with console")
        print(df.show(n=self.limit))
        return df, Shape()

    def outlet(self) -> tuple[str, str]:
        return "console", self.dest()

    def dest(self) -> str:
        return "console"



from typing import Final

from airflow.www.utils import UIAlert

DASHBOARD_UIALERTS: Final[list[UIAlert]] = [
    UIAlert("Welcome to Airflow Standalone Mode", category="warning"),
]
STATE_COLORS: Final[dict[str, str]] = {
    "deferred": "#a9a9a9",  # Dark Gray
    "failed": "#c94c4c",  # Muted Red
    "queued": "#d3d3d3",  # Light Gray
    "removed": "#f5f5f5",  # Lighter Gray
    "restarting": "#7f7f7f",  # Medium Gray
    "running": "#5f9ea0",  # Cadet Blue
    "scheduled": "#f0e68c",  # Khaki
    "shutdown": "#696969",  # Dim Gray
    "skipped": "#b0c4de",  # Light Steel Blue
    "success": "#8fbc8f",  # Dark Sea Green
    "up_for_reschedule": "#add8e6",  # Light Blue
    "up_for_retry": "#ffdead",  # Navajo White
    "upstream_failed": "#ffa07a",  # Light Salmon
}



from typing import Annotated, Union

from pydantic import Field

from .console import Console
from .files.local import LocalCSVFile

Sink = Annotated[
    Union[
        LocalCSVFile,
        Console,
    ],
    Field(discriminator="type"),
]



from __future__ import annotations

import json
import logging
import re
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

from .utils import is_remote_session

logger = logging.getLogger("jett")


def extract_table_names_from_spark_plan(items: Any) -> list[str]:
    """A recursive function for getting table names from a spark plan dictionary.

        For CTE and table names, it will be in the element that has
    `class=org.apache.spark.sql.catalyst.analysis.UnresolvedRelation`

    Args:
        items: dictionary or list containing spark plan data

    Returns:
        list[str]: list of result table names
    """
    tables: list[str] = []
    if isinstance(items, dict):
        if (
            items.get("class")
            == "org.apache.spark.sql.catalyst.analysis.UnresolvedRelation"
        ):
            multipart_identifier: str = items.get("multipartIdentifier", "[]")
            if multipart_identifier:
                table_name: str = ".".join(
                    multipart_identifier.strip("[]").split(", ")
                )
                tables.append(table_name)
        for value in items.values():
            if isinstance(value, (dict, list)):
                tables.extend(extract_table_names_from_spark_plan(value))
    elif isinstance(items, list):
        for item in items:
            tables.extend(extract_table_names_from_spark_plan(item))
    return tables


def extract_table_names_from_query(query: str) -> list[str]:
    """A function that return the inlets table name from SparkSQL query string.

    Args:
        query (str): SparkSQL query string.

    Returns:
        list[str]: list of inlets table names
    """
    # NOTE: Got None for now, So I will use `getOrCreate` first
    # ---
    # spark = SparkSession.getActiveSession()
    spark = SparkSession.builder.appName("sql_parser").getOrCreate()

    if is_remote_session(spark):
        logger.warning(
            "This `extract_table_names_from_query` func is not support for "
            "SparkConnectSession yet"
        )
        return []

    # NOTE: if SparkSession, use Spark Plan to get the table names from query.
    plan = spark._jsparkSession.sessionState().sqlParser().parsePlan(query)
    plan_items: dict[str, Any] | list[Any] = json.loads(plan.toJSON())
    plan_string = plan.toString()
    cte_match = re.match(r"CTE \[(.*?)]", plan_string)
    if cte_match:
        cte_list = [item.strip() for item in cte_match.group(1).split(",")]
    else:
        cte_list = []

    # NOTE: get table names from spark plan
    result_tables: list[str] = extract_table_names_from_spark_plan(plan_items)

    # NOTE: Exclude the CTE alias from extraction result.
    inlet_tables = list(set(result_tables) - set(cte_list))
    inlet_tables.sort()
    return inlet_tables
